# pylint: disable=too-many-lines
# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only



Here we define replacement classes for the case that we have no
python interface to KDE.

"""


import sys
import os
import subprocess
import getpass
import webbrowser
import codecs
import weakref
import urllib
from collections import defaultdict
from typing import List, Dict, TYPE_CHECKING, Optional, Union, Callable, Any, Type, cast

# pylint: disable=wrong-import-order

ParamValue = Union[str, int, bool]

if TYPE_CHECKING:
    from deferredutil import Deferred
    from mainwindow import MainWindow
    from config import SetupPreferences

# pylint: disable=wrong-import-position

from configparser import ConfigParser, NoSectionError, NoOptionError

# here come the replacements:

# pylint: disable=wildcard-import,unused-wildcard-import
from qt import *
from qtpy import QT6, PYSIDE6, QT_VERSION, API_NAME, PYQT_VERSION

# pylint: disable=wrong-import-position

from mi18n import MLocale, KDETranslator, i18n, i18nc

from common import Internal, isAlive, Debug
from util import popenReadlines
from statesaver import StateSaver

if sys.platform != 'win32':
    import pwd

__all__ = ['KApplication', 'KConfig',
           'KMessageBox', 'KConfigSkeleton', 'KDialogButtonBox',
           'KConfigDialog', 'KDialog',
           'KUser', 'KStandardAction',
           'KXmlGuiWindow', 'KGlobal', 'KIcon']


class KApplication(QApplication):

    """stub"""

    def __init__(self) ->None:
        assert not Internal.isServer, 'KApplication is not supported for the server'
        super().__init__(sys.argv)

        # Qt uses sys.argv[0] as application name
        # which is used by QStandardPaths - if we start kajongg.py directly,
        # the search path would look like /usr/share/kajongg.py/

        self.translators:List[KDETranslator] = []
        self.setApplicationName('kajongg')
        self.setApplicationVersion(str(Internal.defaultPort))

        self.initQtTranslator()

    def installTranslatorFile(self, qmName:str) ->None:
        """qmName is a full path to a .qm file"""
        if os.path.exists(qmName):
            translator = KDETranslator(self)
            translator.load(qmName)
            self.installTranslator(translator)
            self.translators.append(translator)
            if Debug.i18n:
                Internal.logger.debug('Installed Qt translator from %s', qmName)


    def initQtTranslator(self) ->None:
        """load translators using Qt .qm files"""
        for _ in self.translators:
            self.removeTranslator(_)
        self.translators = []
        _ = KDETranslator(self)
        self.translators.append(_)
        self.installTranslator(_)
        for language in reversed(list(MLocale.extendRegionLanguages(MLocale.currentLanguages()))):
            _ = QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath)
            self.installTranslatorFile(os.path.join( _, f'qtbase_{language}.qm'))
            self.installTranslatorFile(f'/usr/share/locale/{language}/LC_MESSAGES/kwidgetsaddons5_qt.qm')

    @classmethod
    def desktopSize(cls) -> QSize:
        """The size of the current screen"""
        return QGuiApplication.primaryScreen().availableGeometry()


class CaptionMixin:

    """used by KDialog and KXmlGuiWindow"""

    def setCaption(self, caption:str) ->None:
        """append app name"""
        if caption:
            if not caption.endswith(i18n('Kajongg')):
                caption += f" – {i18n('Kajongg')}"
        else:
            caption = i18n('Kajongg')
        self.setWindowTitle(caption)  # type:ignore
        self.setWindowIcon(KIcon('kajongg'))  # type:ignore


class Help:
    """Interface to the KDE help system"""

    @staticmethod
    def url_for_language(lang:str) ->str:
        """The url for a given language"""
        return f'https://docs.kde.org/stable/{lang}/kajongg/kajongg/index.html'

    @staticmethod
    def find_help_url() ->str:
        """find an existing help url for preferred language"""
        for language in Internal.kajonggrc.group('Locale').readEntry('Language').split(':'):
            try:
                url = Help.url_for_language(language)
                with urllib.request.urlopen(url) as _:
                    if _.status == 200:
                        return url
            except urllib.error.HTTPError:
                pass
        return Help.url_for_language('en')

    @staticmethod
    def start() ->None:
        """start the KDE help center for kajongg or go to docs.kde.org"""
        try:
            subprocess.Popen(['khelpcenter', 'help:/kajongg/index.html'])  # pylint:disable=consider-using-with
        except OSError:
            webbrowser.open(Help.find_help_url())


class IconLabel(QLabel):

    """for use in messages and about dialog"""

    def __init__(self, iconName:str, dialog:QDialog) ->None:
        super().__init__()
        icon = KIcon(iconName)
        option = QStyleOption()
        option.initFrom(dialog)
        style = dialog.style()
        assert style
        self.setPixmap(icon.pixmap(style.pixelMetric(QStyle.PixelMetric.PM_MessageBoxIconSize, option, dialog)))


class KMessageBox:

    """again only what we need"""
    NoExec = 1
    AllowLink = 2
    Options = int

    @staticmethod
    def createKMessageBox(
            dialog:'KDialog', icon:QMessageBox.Icon, text:str, unusedStrlist:List[str],
            unusedAsk:str, unusedCheckboxReturn:bool, options:int) ->None:
        """translated as far as needed from kmessagegox.cpp"""
        mainLayout = QVBoxLayout()

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        hLayout.setSpacing(-1)
        mainLayout.addLayout(hLayout, 5)

        iconName = {
            QMessageBox.Icon.Information: 'dialog-information',
            QMessageBox.Icon.Warning: 'dialog-warning',
            QMessageBox.Icon.Question: 'dialog-question'}[icon]
        iconLayout = QVBoxLayout()
        iconLayout.addStretch(1)
        iconLayout.addWidget(IconLabel(iconName, dialog))
        iconLayout.addStretch(5)
        hLayout.addLayout(iconLayout, 0)

        messageLabel = QLabel(text)
        flags = Qt.TextInteractionFlag.TextSelectableByMouse
        if options & KMessageBox.AllowLink:
            flags = cast(Qt.TextInteractionFlag, flags | Qt.TextInteractionFlag.LinksAccessibleByMouse)
            messageLabel.setOpenExternalLinks(True)
        messageLabel.setTextInteractionFlags(flags)

        desktop = KApplication.desktopSize()
        if messageLabel.sizeHint().width() > desktop.width() * 0.5:
            messageLabel.setWordWrap(True)

        usingScrollArea = desktop.height(
        ) // 3 < messageLabel.sizeHint(
        ).height(
        )
        if usingScrollArea:
            scrollArea = QScrollArea(dialog)
            scrollArea.setWidget(messageLabel)
            scrollArea.setWidgetResizable(True)
            scrollArea.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            scrollPal = QPalette(scrollArea.palette())
            viewport = scrollArea.viewport()
            if viewport:
                viewport.setPalette(scrollPal)
            hLayout.addWidget(scrollArea, 5)
        else:
            hLayout.addWidget(messageLabel, 5)

        mainLayout.addWidget(dialog.buttonBox)
        dialog.setLayout(mainLayout)

KDialogButtonBox = QDialogButtonBox


class KDialog(CaptionMixin, QDialog):

    """QDialog should be enough for kajongg"""
    NoButton = QDialogButtonBox.StandardButton.NoButton
    Ok = QDialogButtonBox.StandardButton.Ok
    Cancel = QDialogButtonBox.StandardButton.Cancel
    Yes = QDialogButtonBox.StandardButton.Yes
    No = QDialogButtonBox.StandardButton.No
    Help = QDialogButtonBox.StandardButton.Help
    Apply = QDialogButtonBox.StandardButton.Apply
    RestoreDefaults = QDialogButtonBox.StandardButton.RestoreDefaults
    Default = QDialogButtonBox.StandardButton.RestoreDefaults
    Close = QDialogButtonBox.StandardButton.Close

    def __init__(self, parent:Optional[QWidget]=None) ->None:
        super().__init__(parent)
        self.buttonBox = QDialogButtonBox()
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.__mainWidget:Optional[QWidget] = None

    def __set_text(self, buttonMask: QDialogButtonBox.StandardButton,
        which:QDialogButtonBox.StandardButton, text:str) ->None:
        """set button text"""
        if which & buttonMask:
            button = self.buttonBox.button(which)
            if button:
                button.setText(text)

    def __connect_button(self, buttonMask: QDialogButtonBox.StandardButton,
        which:QDialogButtonBox.StandardButton, connector:Any) ->None:
        """set connector"""
        if which & buttonMask:
            button = self.buttonBox.button(which)
            if button:
                button.clicked.connect(connector)

    def setButtons(self, buttonMask:QDialogButtonBox.StandardButton) ->None:
        """(re)create the buttonbox and put all wanted buttons into it"""
        if not buttonMask:
            self.buttonBox.clear()
            return
        self.buttonBox.setStandardButtons(buttonMask)
        self.__set_text(buttonMask, KDialog.Ok, i18n('&OK'))
        self.__set_text(buttonMask, KDialog.Apply, i18n('&Apply'))
        self.__set_text(buttonMask, KDialog.Cancel, i18n('&Cancel'))
        self.__set_text(buttonMask, KDialog.Help, i18n('&Help'))
        self.__set_text(buttonMask, KDialog.RestoreDefaults, i18n('&Defaults'))
        self.__connect_button(buttonMask, KDialog.RestoreDefaults, self.restoreDefaults)
        self.__connect_button(buttonMask, KDialog.Help, Help.start)

    def restoreDefaults(self) ->None:
        """virtual"""

    def setMainWidget(self, widget:QWidget) ->None:
        """see KDialog.setMainWidget"""
        if self.layout() is None:
            QVBoxLayout(self)
            _ = self.layout()
            assert _ is not None
            _.addWidget(self.buttonBox)
        layout = self.layout()
        assert layout is not None
        if self.__mainWidget:
            layout.removeWidget(self.__mainWidget)
            layout.removeWidget(self.buttonBox)
        self.__mainWidget = widget
        layout.addWidget(widget)
        layout.addWidget(self.buttonBox)

    def button(self, buttonCode:QDialogButtonBox.StandardButton) ->QPushButton:
        """return the matching button"""
        result = self.buttonBox.button(buttonCode)
        assert result
        return result

    def returns(self, button:Optional['QDialogButtonBox.StandardButton']=None) ->Any:
        """the user answered"""
        if button is None:
            button = self.default  # type:ignore[attr-defined]
        return button in (KDialog.Yes, KDialog.Ok)


class KUser:

    """only the things kajongg needs"""

    def __init__(self, uid:Optional[int]=None) ->None:
        pass

    def fullName(self) ->str:
        """stub"""
        if sys.platform == 'win32':
            return self.loginName()
        return pwd.getpwnam(self.loginName()).pw_gecos.replace(',', '')  # pylint:disable=possibly-used-before-assignment

    @staticmethod
    def loginName() ->str:
        """stub"""
        return getpass.getuser()


class KStandardAction:

    """stub"""
    @classmethod
    def preferences(cls, slot:Callable, actionCollection:'KActionCollection') ->None:
        """should add config dialog menu entry"""
        mainWindow = Internal.mainWindow
        assert mainWindow
        separator = Action(mainWindow)
        separator.setSeparator(True)
        mainWindow.actionStatusBar = Action(mainWindow, 'options_show_statusbar', None)
        mainWindow.actionStatusBar.setCheckable(True)
        mainWindow.actionStatusBar.setEnabled(True)
        mainWindow.actionStatusBar.toggled.connect(mainWindow.toggleStatusBar)
        mainWindow.actionStatusBar.setText(
            i18nc('@action:inmenu', "Show St&atusbar"))
        mainWindow.actionToolBar = Action(mainWindow, 'options_show_toolbar', None)
        mainWindow.actionToolBar.setCheckable(True)
        mainWindow.actionToolBar.setEnabled(True)
        mainWindow.actionToolBar.toggled.connect(mainWindow.toggleToolBar)
        mainWindow.actionToolBar.setText(
            i18nc('@action:inmenu', "Show &Toolbar"))

        actionCollection.addAction('', separator)

        action = Action(mainWindow)
        action.triggered.connect(mainWindow.configureToolBar)
        action.setText(i18n('Configure Tool&bars...'))
        action.setIcon(KIcon('configure-toolbars'))
        action.setIconText(i18n('Configure toolbars'))
        separator = Action(mainWindow)
        separator.setSeparator(True)
        actionCollection.addAction('options_configure_toolbars', action)

        action = Action(mainWindow)
        action.triggered.connect(slot)
        action.setText(i18n('Configure &Kajongg...'))
        action.setIcon(KIcon('configure'))
        action.setIconText(i18n('Configure'))
        actionCollection.addAction('options_configure', action)


class KActionCollection:

    """stub"""

    def __init__(self, mainWindow:'KXmlGuiWindow') ->None:
        self.__actions:Dict[str,'Action'] = {}
        self.mainWindow = mainWindow

    def addAction(self, name:str, action:'Action') ->None:
        """stub"""
        self.__actions[name] = action
        for content in self.mainWindow.menus.values():
            if name in content[1]:
                content[0].addAction(action)
                break

    def actions(self) ->Dict[str, 'Action']:
        """the actions in this collection"""
        return self.__actions


class MyStatusBarItem:

    """one of the four player items"""

    def __init__(self, text:str, idx:int, stretch:int=0) ->None:
        self.idx = idx
        self.stretch = stretch
        self.label = QLabel()
        self.label.setText(text)
        self.label.setAlignment(cast(Qt.AlignmentFlag, Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter))


class KStatusBar(QStatusBar):

    """stub"""

    def __init__(self, *args:Any, **kwargs:Any) ->None:
        super().__init__(*args, **kwargs)
        self.__items:List[MyStatusBarItem] = []

    def hasItem(self, idx:int) ->bool:
        """stub"""
        return len(self.__items) > idx

    def changeItem(self, text:str, idx:int) ->None:
        """stub"""
        self.__items[idx].label.setText(text)

    def insertItem(self, text:str, idx:int, stretch:int) ->None:
        """stub"""
        item = MyStatusBarItem(text, idx, stretch)
        self.__items.append(item)
        self.insertWidget(item.idx, item.label, stretch=item.stretch)

    def removeItem(self, idx:int) ->None:
        """stub"""
        item = self.__items[idx]
        self.removeWidget(item.label)
        del self.__items[idx]

    def setItemAlignment(self, idx:int, alignment:Qt.AlignmentFlag) ->None:
        """stub"""
        self.__items[idx].label.setAlignment(alignment)


class KXmlGuiWindow(CaptionMixin, QMainWindow):

    """stub"""

    def __init__(self) ->None:
        super().__init__()
        self.actionStatusBar: 'Action'
        self.actionToolBar: 'Action'
        self.actionFullscreen: 'Action'
        self._actions = KActionCollection(self)
        self._toolBar = QToolBar(self)
        self._toolBar.setObjectName('Toolbar')
        self.addToolBar(self.toolBar())
        self.setStatusBar(KStatusBar(self))
        self.statusBar().setObjectName('StatusBar')
        self.menus = {}
        # the menuItems are added to the main  menu by KActionCollection.addAction
        # their order is not defined by this here but by the order in which MainWindow
        # creates the menu entries. This only defines which action goes into which main menu.
        for menu, menuItems in (
                (i18n('&Game'), ('scoreGame', 'play', 'abort', 'quit')),
                (i18n('&View'), ('scoreTable', 'explain', 'chat', 'fullscreen')),
                (i18n('&Settings'), ('players', 'rulesets', 'angle', 'demoMode', '', 'options_show_statusbar',
                                     'options_show_toolbar', '', 'options_configure_toolbars', 'options_configure')),
                (i18n('&Help'), ('help', 'language', 'aboutkajongg'))):
            mainMenu = QMenu(menu)
            self.menus[menu] = (mainMenu, menuItems)
            menu_bar = self.menuBar()
            if menu_bar:
                menu_bar.addMenu(mainMenu)
        self.setCaption('')
        self.actionHelp = Action(self, "help", "help-contents", Help.start)
        self.actionHelp.setText(i18nc('@action:inmenu', '&Help'))
        self.actionLanguage = Action(self,
            "language", "preferences-desktop-locale", self.selectLanguage)
        self.actionLanguage.setText(i18n('Switch Application Language'))
        self.actionAboutKajongg = Action(self,
            'aboutkajongg', 'kajongg', self.aboutKajongg)
        self.actionAboutKajongg.setText(
            i18nc('@action:inmenu', 'About &Kajongg'))
        self.toolBar().setMovable(False)
        self.toolBar().setFloatable(False)
        self.toolBar().setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)

    def statusBar(self) -> KStatusBar:
        """for mypy"""
        return cast(KStatusBar, QMainWindow.statusBar(self))

    def showEvent(self, event:Optional[QShowEvent]) ->None:
        """now that the MainWindow code has run, we know all actions"""
        self.refreshToolBar()
        assert Internal.Preferences
        self.toolBar().setVisible(bool(Internal.Preferences.toolBarVisible))
        self.actionStatusBar.setChecked(self.statusBar().isVisible())
        self.actionToolBar.setChecked(self.toolBar().isVisible())
        self.actionFullscreen.setChecked(
            self.windowState() & Qt.WindowState.WindowFullScreen == Qt.WindowState.WindowFullScreen)
        if event:
            super().showEvent(event)

    def hideEvent(self, event:Optional[QHideEvent]) ->None:
        """save status"""
        assert Internal.Preferences
        Internal.Preferences.toolBarVisible = self.toolBar(
        ).isVisible()
        if event:
            super().hideEvent(event)

    def toggleStatusBar(self, checked:bool) ->None:
        """show / hide status bar"""
        self.statusBar().setVisible(checked)

    def toggleToolBar(self, checked:bool) ->None:
        """show / hide status bar"""
        self.toolBar().setVisible(checked)

    def configureToolBar(self) ->None:
        """configure toolbar"""
        dlg = KEditToolBar(self)
        dlg.show()

    def refreshToolBar(self) ->None:
        """reload settings for toolbar actions"""
        self.toolBar().clear()
        assert Internal.Preferences
        for name in Internal.Preferences.toolBarActions.split(','):
            self.toolBar().addAction(self.actionCollection().actions()[name])

    def actionCollection(self) ->'KActionCollection':
        """stub"""
        return self._actions

    def setupGUI(self) ->None:
        """stub"""

    def toolBar(self) ->QToolBar:
        """stub"""
        return self._toolBar

    @staticmethod
    def selectLanguage() ->None:
        """switch the language"""
        assert Internal.mainWindow
        KSwitchLanguageDialog(Internal.mainWindow).exec()

    @staticmethod
    def aboutKajongg() ->None:
        """show an about dialog"""
        assert Internal.mainWindow
        AboutKajonggDialog(Internal.mainWindow).exec()

    def queryClose(self) ->bool:
        """default"""
        return True

    def queryExit(self) ->bool:
        """default"""
        return True

    def closeEvent(self, event:Optional[QEvent]) ->None:
        """call queryClose/queryExit"""
        if event:
            if self.queryClose() and self.queryExit():
                event.accept()
            else:
                event.ignore()


class KConfigGroup:

    """mimic KConfigGroup as far as we need it"""

    def __init__(self, config:'KConfig', groupName:str) ->None:
        self.config = weakref.ref(config)
        self.groupName = groupName

    def __default(self, name:str) ->Optional[ParamValue]:
        """defer computation of Languages until really needed"""
        if self.groupName == 'Locale' and name == 'Language':
            return QLocale().name()
        return None

    def readEntry(self, name:str) ->Optional[ParamValue]:
        """get an entry from this group."""
        try:
            _ = self.config()
            assert _
            items = _.items(self.groupName)
        except NoSectionError:
            return self.__default(name)
        items = {x: y for x, y in items if x.startswith(name)}
        i18nItems = {x: y for x, y in items.items() if x.startswith(name + '[')}
        if i18nItems:
            languages = Internal.kajonggrc.group('Locale').readEntry('Language').split(':')
            languages = [x.split('_')[0] for x in languages]
            for language in languages:
                key = f'{name}[{language}]'
                if key in i18nItems:
                    return i18nItems[key]
        if name in items:
            if self.groupName == 'Locale' and name == 'Language':
                languages = [x for x in items[name].split(':') if MLocale.isLanguageInstalled(x)]
                if languages:
                    return ':'.join(languages)
                return QLocale().name()
            return items[name]
        return self.__default(name)

    def readInteger(self, name:str) ->int:
        """calls readEntry and returns it as an int or raises an Exception."""
        result = self.readEntry(name)
        if result is None:
            raise ValueError(f'{self}: cannot find {name}')
        try:
            return int(result)
        except ValueError as _:
            raise ValueError(f'{self}: cannot parse {name}') from _

    def __str__(self) ->str:
        config = self.config()
        path = config.path if config else 'NO_PATH'
        return f'{path}:[{self.groupName}]'


class KGlobal:

    """stub"""

    @classmethod
    def initStatic(cls) ->None:
        """init class members"""
        Internal.kajonggrc = KConfig()

class KConfig(ConfigParser):

    """Parse KDE config files.
    This mimics KDE KConfig but can also be used like any ConfigParser but
    without support for a default section."""

    def __init__(self, path:Optional[str]=None) ->None:
        super().__init__(delimiters=('=', ))
        if path is None:
            path = QStandardPaths.writableLocation(QStandardPaths.StandardLocation.AppConfigLocation)
            path = os.path.join(path, 'kajonggrc')
        self.path = os.path.expanduser(path)
        if os.path.exists(self.path):
            with codecs.open(self.path, 'r', encoding='utf-8') as cfgFile:
                self.read_file(cfgFile)

    def optionxform(self, optionstr:str) ->str:
        """KDE needs upper/lowercase distinction"""
        return optionstr

    def setValue(self, section:str, option:str, value:int) ->None:
        """like set but add missing section"""
        if section not in self.sections():
            self.add_section(section)
        self.set(section, option, str(value))

    def writeToFile(self) ->None:
        """Write an .ini-format representation of the configuration state."""
        with open(self.path, 'w', encoding='utf-8') as filePointer:
            self.write(filePointer, space_around_delimiters=False)

    def group(self, groupName:str) ->KConfigGroup:
        """just like KConfig"""
        return KConfigGroup(self, groupName)

def KIcon(name:Optional[str]=None) ->QIcon:  # pylint: disable=invalid-name
    """simple wrapper"""
    if sys.platform == 'win32':
        return QIcon(os.path.join('share', 'icons', name) if name else None)
    if name:
        return QIcon.fromTheme(name)
    return QIcon()


class Action(QAction):

    """helper for creation QAction"""

    def __init__(self, parent:'KXmlGuiWindow', name:Optional[str]=None,
        icon:Optional[str]=None, slot:Optional[Callable]=None,
        shortcut:Optional[str]=None, actionData:Optional[Union[str, Type[QWidget]]]=None) ->None:
        super().__init__(parent)
        if icon:
            self.setIcon(KIcon(icon))
        if slot:
            self.triggered.connect(slot)
        if parent:
            if name is None:
                name = ''
            parent.actionCollection().addAction(name, self)
        if shortcut:
            self.setShortcut(QKeySequence(shortcut))
            self.setShortcutContext(Qt.ShortcutContext.ApplicationShortcut)
        if actionData is not None:
            self.setData(actionData)


class KConfigSkeletonItem:

    """one preferences setting used by KOnfigSkeleton"""

    def __init__(self, skeleton:'KConfigSkeleton', key:str, value:ParamValue,
        default:ParamValue) ->None:
        assert skeleton
        self.skeleton = skeleton
        self.group = skeleton.currentGroup
        self.key = key
        self._value = value
        if value is None:
            self._value = default
        self.default = default

    def value(self) ->ParamValue:
        """default getter"""
        return self._value

    def setValue(self, value:ParamValue) ->None:
        """default setter"""
        if self._value != value:
            self._value = value
            self.skeleton.configChanged.emit()

    def getFromConfig(self) ->None:
        """if not there, use default"""
        try:
            self._value = Internal.kajonggrc.get(self.group, self.key)
        except (NoSectionError, NoOptionError):
            self._value = self.default  # type:ignore


class ItemBool(KConfigSkeletonItem):

    """boolean preferences setting used by KOnfigSkeleton"""

    def __init__(self, skeleton:'KConfigSkeleton', key:str, value:bool, default:bool) ->None:
        super().__init__(skeleton, key, value, default)

    def getFromConfig(self) ->None:
        """if not there, use default"""
        try:
            self._value = Internal.kajonggrc.getboolean(self.group, self.key)
        except (NoSectionError, NoOptionError):
            self._value = self.default


class ItemString(KConfigSkeletonItem):

    """string preferences setting used by KOnfigSkeleton"""

    def __init__(self, skeleton:'KConfigSkeleton', key:str, value:str, default:str) ->None:
        if value == '':
            value = default
        super().__init__(skeleton, key, value, default)


class ItemInt(KConfigSkeletonItem):

    """integer preferences setting used by KOnfigSkeleton"""

    def __init__(self, skeleton:'KConfigSkeleton', key:str, value:int, default:int) ->None:
        super().__init__(skeleton, key, value, default)
        self.minValue = -99999
        self.maxValue = 99999999

    def getFromConfig(self) ->None:
        """if not there, use default"""
        try:
            self._value = Internal.kajonggrc.getint(self.group, self.key)
        except (NoSectionError, NoOptionError):
            self._value = self.default

    def setMinValue(self, value:int) ->None:
        """minimum value for this setting"""
        self.minValue = value

    def setMaxValue(self, value:int) ->None:
        """maximum value for this setting"""
        self.maxValue = value


class KConfigSkeleton(QObject):

    """handles preferences settings"""
    configChanged = Signal()

    def __init__(self) ->None:
        super().__init__()
        self.currentGroup:str
        self.items:List[KConfigSkeletonItem] = []
        self.toolBarActions:str
        self.addBool('MainWindow', 'toolBarVisible', True)
        self.addString(
            'MainWindow',
            'toolBarActions',
            'quit,play,scoreTable,explain,players,options_configure')

    def addBool(self, group:str, name:str, default:Optional[bool]=None) ->None:
        """to be overridden"""

    def addString(self, group:str, name:str, default:Optional[str]=None) ->None:
        """to be overridden"""

    def readConfig(self) ->None:
        """init already read config"""
        for item in self.items:
            item.getFromConfig()

    def writeConfig(self) ->None:
        """to the same file name"""
        for item in self.items:
            Internal.kajonggrc.setValue(item.group, item.key, item.value())
        Internal.kajonggrc.writeToFile()

    def as_dict(self) ->defaultdict:
        """a dict of dicts"""
        result: defaultdict = defaultdict(dict)
        for item in self.items:
            result[item.group][item.key] = item.value()
        return result

    def setCurrentGroup(self, group:str) ->None:
        """to be used by following add* calls"""
        self.currentGroup = group

    def addItem(self, key:str, value:ParamValue,
        default:ParamValue) ->KConfigSkeletonItem:
        """add a string preference"""
        cls: Type[KConfigSkeletonItem]
        if isinstance(value, bool):
            cls = ItemBool
        elif isinstance(value, int):
            cls = ItemInt
        elif isinstance(value, str):
            cls = ItemString
        else:
            raise TypeError(f'addiItem accepts only bool, int, str but not {type(value)}/{value}')
        result = cls(self, key, value, default)
        result.getFromConfig()
        self.items.append(result)
        return result


class KSwitchLanguageDialog(KDialog):
    """select application language"""

    def __init__(self, parent:QWidget) ->None:
        super().__init__(parent)
        self.languageRows:Dict[QWidget, Any] = {}
        self.languageButtons:List[KLanguageButton] = []
        self.setCaption(i18n('Switch Application Language'))
        self.widget = QWidget()
        topLayout = QVBoxLayout()
        self.widget.setLayout(topLayout)

        topLayout.addWidget(QLabel(i18n("Please choose the language which should be used for this application:")))

        languageHorizontalLayout = QHBoxLayout()
        topLayout.addLayout(languageHorizontalLayout)

        self.languagesLayout = QGridLayout()
        languageHorizontalLayout.addLayout(self.languagesLayout)
        languageHorizontalLayout.addStretch()

        defined = Internal.kajonggrc.group(
            'Locale').readEntry('Language').split(':')
        if not defined:
            defined = [QLocale().name()]
        for idx, _ in enumerate(defined):
            self.addLanguageButton(_, isPrimaryLanguage=idx == 0)

        addButtonHorizontalLayout = QHBoxLayout()
        topLayout.addLayout(addButtonHorizontalLayout)

        addLangButton = QPushButton(i18n("Add Fallback Language"), self)
        addLangButton.clicked.connect(self.slotAddLanguageButton)
        addButtonHorizontalLayout.addWidget(addLangButton)
        addButtonHorizontalLayout.addStretch()

        topLayout.addStretch(10)

        self.setButtons(KDialog.Ok | KDialog.Cancel | KDialog.RestoreDefaults)
        self.setMainWidget(self.widget)

    def addLanguageButton(self, languageCode: str, isPrimaryLanguage: bool) ->None:
        """add button for language"""
        labelText = i18n("Primary language:") if isPrimaryLanguage else i18n("Fallback language:")
        languageButton = KLanguageButton('', self.widget)
        languageButton.current = languageCode

        removeButton = None
        if not isPrimaryLanguage:
            removeButton = QPushButton(i18n("Remove"))
            removeButton.clicked.connect(self.removeButtonClicked)

        languageButton.setToolTip(
            i18n("This is the main application language which will be used first, before any other languages.")
            if isPrimaryLanguage else
            i18n("This is the language which will be used if any previous "
                 "languages do not contain a proper translation."))

        numRows = self.languagesLayout.rowCount()

        languageLabel = QLabel(labelText)
        self.languagesLayout.addWidget(languageLabel, numRows + 1, 1, Qt.AlignmentFlag.AlignLeft)
        self.languagesLayout.addWidget(languageButton.button, numRows + 1, 2, Qt.AlignmentFlag.AlignLeft)

        if not isPrimaryLanguage:
            assert removeButton
            self.languagesLayout.addWidget(removeButton, numRows + 1, 3, Qt.AlignmentFlag.AlignLeft)
            removeButton.show()
            self.languageRows[removeButton] = tuple([languageLabel, languageButton])

        self.languageRows[languageButton] = tuple([languageLabel, languageButton])
        self.languageButtons.append(languageButton)

    def accept(self) ->None:
        """OK"""
        newValue = ':'.join(x.current for x in self.languageButtons)
        Internal.kajonggrc.setValue('Locale', 'Language', newValue)
        super().accept()

    def slotAddLanguageButton(self) ->None:
        """adding a new button with en_US as it should always be present"""
        self.addLanguageButton('en_US', len(self.languageButtons) == 0)

    def restoreDefaults(self) ->None:
        """reset values to default"""
        for _ in self.languageRows:
            if isinstance(_, KLanguageButton):
                self.removeLanguage(_)
        for removeButton in self.languageRows:
            if isAlive(removeButton):
                removeButton.deleteLater()
        self.languageRows = {}
        self.addLanguageButton(QLocale().name(), True)

    def removeButtonClicked(self) ->None:
        """remove this language"""
        _ = self.sender()
        assert isinstance(_, KLanguageButton)
        self.removeLanguage(_)

    def removeLanguage(self, button:'KLanguageButton') ->None:
        """remove this language"""
        label, languageButton = self.languageRows[button]
        label.deleteLater()
        languageButton.deleteLater()
        button.deleteLater()
        del self.languageRows[languageButton]
        self.languageButtons.remove(languageButton)


class KLanguageButton(QWidget):
    """A language button for KSwitchLanguageDialog"""

    def __init__(self, txt:str, parent:QWidget) ->None:
        super().__init__(parent)
        self.button = QPushButton(txt)
        self.popup = QMenu()
        self.button.setMenu(self.popup)
        self.setText(txt)
        self.__currentItem:str
        for _ in MLocale.availableLanguages_().split(':'):
            self.addLanguage(_)
        self.popup.triggered.connect(self.slotTriggered)
        self.popup.show()
        self.button.show()
        self.show()

    def deleteLater(self) ->None:
        """self and children"""
        self.button.deleteLater()
        self.popup.deleteLater()
        super().deleteLater()

    def setText(self, txt:str) ->None:
        """proxy: sets the button text"""
        self.button.setText(txt)


    def addLanguage(self, languageCode:str) ->None:
        """add language to popup"""
        text = languageCode
        locale = QLocale(languageCode)
        if locale != QLocale.c():
            text = locale.nativeLanguageName()

        action = QAction(QIcon(), text, self)
        action.setData(languageCode)
        self.popup.addAction(action)

    @property
    def current(self) ->str:
        """current languageCode"""
        return self.__currentItem

    @current.setter
    def current(self, languageCode:str) ->None:
        """point to languageCode"""
        action = (
            self.findAction(languageCode)
            or self.findAction(languageCode.split('_')[0])
            or self.popup.actions()[0])
        self.__currentItem = action.data()
        self.button.setText(action.text())

    def slotTriggered(self, action:Action) ->None:
        """another language has been selected from the popup"""
        self.current = action.data()

    def findAction(self, data:str) ->Optional[QAction]:
        """find action by name"""
        for action in self.popup.actions():
            if action.data() == data:
                assert isinstance(action, QAction)
                return action
        return None


class AboutKajonggDialog(KDialog):
    """about kajongg dialog"""

    def __init__(self, parent:QWidget) ->None:
        # pylint: disable=too-many-locals, too-many-statements
        from twisted import __version__

        super().__init__(parent)
        self.setCaption(i18n('About Kajongg'))
        self.setButtons(KDialog.Close)
        vLayout = QVBoxLayout()
        hLayout1 = QHBoxLayout()
        hLayout1.addWidget(IconLabel('kajongg', self))
        h1vLayout = QVBoxLayout()
        h1vLayout.addWidget(QLabel('Kajongg'))
        try:
            from appversion import VERSION  # type:ignore[import]
        except ImportError:
            VERSION = "Unknown"

        assert isinstance(QT_VERSION, str)
        underVersions = ['Qt' + QT_VERSION +' API=' + API_NAME]
        if PYQT_VERSION:
            underVersions.append('sip ' + SIP_VERSION_STR)
        if PYSIDE6:
            import PySide6
            underVersions.append('PySide6 ' + PySide6.__version__)

        h1vLayout.addWidget(QLabel(i18n('Version: %1', VERSION)))
        h1vLayout.addWidget(QLabel(i18n('Protocol version %1', Internal.defaultPort)))
        authors = ((
            "Wolfgang Rohdewald",
            i18n("Original author"),
            "wolfgang@rohdewald.de"), )
        try:
            versions = popenReadlines('kf5-config -v')
            versionsDict = dict(x.split(': ') for x in versions if ':' in x)
            underVersions.append(f"KDE Frameworks {versionsDict['KDE Frameworks']}")
        except OSError:
            underVersions.append(i18n('KDE Frameworks (not installed or not usable)'))
        underVersions.append(f'Twisted {__version__}')
        underVersions.append('Python {}.{}.{} {}'.format(*sys.version_info[:5]))  # pylint:disable=consider-using-f-string
        h1vLayout.addWidget(
            QLabel(
                i18nc('kajongg',
                      'Using versions %1',
                      ', '.join(
                          underVersions))))
        hLayout1.addLayout(h1vLayout)
        spacerItem = QSpacerItem(
            20,
            20,
            QSizePolicy.Expanding,
            QSizePolicy.Expanding)
        hLayout1.addItem(spacerItem)
        vLayout.addLayout(hLayout1)
        tabWidget = QTabWidget()

        aboutWidget = QWidget()
        aboutLayout = QVBoxLayout()
        aboutLabel = QLabel()
        aboutLabel.setWordWrap(True)
        aboutLabel.setOpenExternalLinks(True)
        aboutLabel.setText(
            '<br /><br />'.join([
                i18n("Mah Jongg - the ancient Chinese board game for 4 players"),
                i18n("This is the classical Mah Jongg for four players. "
                     "If you are looking for Mah Jongg solitaire please "
                     "use the application kmahjongg."),
                "(C) 2008-2017 Wolfgang Rohdewald",
                '<a href="https://apps.kde.org/kajongg">https://apps.kde.org/kajongg</a>']))
        licenseLabel = QLabel()
        licenseLabel.setOpenExternalLinks(True)
        licenseLabel.setText(
            '<a href="https://spdx.org/licenses/GPL-2.0-only.html">GNU General Public License Version 2</a>')
        aboutLayout.addWidget(aboutLabel)
        aboutLayout.addWidget(licenseLabel)
        aboutWidget.setLayout(aboutLayout)
        tabWidget.addTab(aboutWidget, '&About')

        authorWidget = QWidget()
        authorLayout = QVBoxLayout()
        bugsLabel = QLabel(
            i18n('Please use <a href="https://bugs.kde.org">https://bugs.kde.org</a> to report bugs.'))
        bugsLabel.setContentsMargins(0, 2, 0, 4)
        bugsLabel.setOpenExternalLinks(True)
        authorLayout.addWidget(bugsLabel)

        titleLabel = QLabel(i18n('Authors:'))
        authorLayout.addWidget(titleLabel)

        for name, description, mail in authors:
            label = QLabel(f'{name} <a href="mailto:{mail}">{mail}</a>: {description}')
            label.setOpenExternalLinks(True)
            authorLayout.addWidget(label)
        spacerItem = QSpacerItem(
            20,
            20,
            QSizePolicy.Expanding,
            QSizePolicy.Expanding)
        authorLayout.addItem(spacerItem)
        authorWidget.setLayout(authorLayout)
        tabWidget.addTab(authorWidget, 'A&uthor')
        vLayout.addWidget(tabWidget)

        vLayout.addWidget(self.buttonBox)
        self.setLayout(vLayout)
        self.buttonBox.setFocus()


class KConfigDialog(KDialog):

    """for the game preferences"""
    dialog = None
    getFunc = {
        'QCheckBox': 'isChecked',
        'QSlider': 'value',
        'QLineEdit': 'text'}
    setFunc = {
        'QCheckBox': 'setChecked',
        'QSlider': 'setValue',
        'QLineEdit': 'setText'}

    def __init__(self, parent:QWidget, name:str, preferences:'SetupPreferences') ->None:
        super().__init__(parent)
        self.pages: List[QWidget]
        self.setCaption(i18n('Configure'))
        self.name = name
        self.preferences = preferences
        self.orgPref:defaultdict[Any, Any]
        self.configWidgets:Dict[str, QWidget] = {}
        self.iconList = QListWidget()
        self.iconList.setViewMode(QListWidget.ViewMode.IconMode)
        self.iconList.setFlow(QListWidget.Flow.TopToBottom)
        self.iconList.setUniformItemSizes(True)
        self.iconList.itemClicked.connect(self.iconClicked)
        self.iconList.currentItemChanged.connect(self.iconClicked)
        self.tabSpace = QStackedWidget()
        self.setButtons(KDialog.Help | KDialog.Ok |
                        KDialog.Apply | KDialog.Cancel | KDialog.RestoreDefaults)
        if button := self.buttonBox.button(KDialog.Apply):
            button.clicked.connect(self.applySettings)
        cmdLayout = QHBoxLayout()
        cmdLayout.addWidget(self.buttonBox)
        self.contentLayout = QHBoxLayout()
        self.contentLayout.addWidget(self.iconList)
        self.contentLayout.addWidget(self.tabSpace)
        layout = QVBoxLayout()
        layout.addLayout(self.contentLayout)
        layout.addLayout(cmdLayout)
        self.setLayout(layout)

    @classmethod
    def showDialog(cls, settings:str) ->None:
        """constructor"""
        assert settings == 'settings'
        if cls.dialog:
            cls.dialog.updateWidgets()
            cls.dialog.updateButtons()
            cls.dialog.show()
            return cls.dialog
        return None

    def showEvent(self, unusedEvent:Optional[QEvent]) ->None:
        """if the settings dialog shows, remember current values
        and show them in the widgets"""
        self.orgPref = self.preferences.as_dict() # FIXME: unused
        self.updateWidgets()

    def iconClicked(self, item:QListWidgetItem) ->None:
        """show the wanted config tab"""
        self.setCurrentPage(
            self.pages[self.iconList.indexFromItem(item).row()])

    @classmethod
    def allChildren(cls, widget:QObject) ->List[QObject]:
        """recursively find all widgets holding settings: Their object name
        starts with kcfg_"""
        result:List[QObject] = []
        for child in widget.children():
            assert isinstance(child, QObject), f'child is:{type(child)}'
            if child.objectName().startswith('kcfg_'):
                result.append(child)
            else:
                result.extend(cls.allChildren(child))
        return result

    def addPage(self, configTab:QWidget, name:str, iconName:str) ->QWidget:
        """add a page to the config dialog"""
        item = QListWidgetItem(KIcon(iconName), name)
        item.setTextAlignment(Qt.AlignmentFlag.AlignHCenter)
        font = item.font()
        font.setBold(True)
        item.setFont(font)
        item.setFlags(cast(Qt.ItemFlag, Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled))
        self.iconList.addItem(item)
        self.tabSpace.addWidget(configTab)
        self.iconList.setIconSize(QSize(80, 80))
        icons = [self.iconList.item(x) for x in range(self.iconList.count())]
        neededIconWidth = max(self.iconList.visualItemRect(x).width()
                              for x in icons)
        margins = self.iconList.contentsMargins()
        neededIconWidth += margins.left() + margins.right()
        self.iconList.setFixedWidth(neededIconWidth)
        self.iconList.setMinimumHeight(120 * self.iconList.count())
        for child in self.allChildren(self):
            assert isinstance(child, QWidget)
            self.configWidgets[
                child.objectName().replace('kcfg_', '')] = child
            if isinstance(child, QCheckBox):
                child.stateChanged.connect(self.updateButtons)  # type:ignore[arg-type]
            elif isinstance(child, QSlider):
                child.valueChanged.connect(self.updateButtons)  # type:ignore[arg-type]
            elif isinstance(child, QLineEdit):
                child.textChanged.connect(self.updateButtons)  # type:ignore[arg-type]
        self.updateButtons()
        return configTab

    def applySettings(self) ->None:
        """Apply pressed"""
        if self.updateButtons():
            self.updateSettings()
            self.updateButtons()

    def accept(self) ->None:
        """OK pressed"""
        if self.updateButtons():
            self.updateSettings()
        super().accept()

    def updateButtons(self) ->bool:
        """Updates the Apply and Default buttons. Returns True if there was a changed setting"""
        changed = False
        for name, widget in self.configWidgets.items():
            oldValue = self.preferences[name]
            newValue = getattr(
                widget,
                self.getFunc[widget.__class__.__name__])()
            if oldValue != newValue:
                changed = True
                break
        if button := self.buttonBox.button(KDialog.Apply):
            button.setEnabled(changed)
        return changed

    def updateSettings(self) ->None:
        """Update the settings from the dialog"""
        for name, widget in self.configWidgets.items():
            self.preferences[name] = getattr(
                widget,
                self.getFunc[widget.__class__.__name__])()
        self.preferences.writeConfig()

    def updateWidgets(self) ->None:
        """Update the dialog based on the settings"""
        self.preferences.readConfig()
        for name, widget in self.configWidgets.items():
            getattr(widget, self.setFunc[widget.__class__.__name__])(
                getattr(self.preferences, name))

    def setCurrentPage(self, page:QWidget) ->None:
        """show wanted page and select its icon"""
        self.tabSpace.setCurrentWidget(page)
        for idx in range(self.tabSpace.count()):
            if item := self.iconList.item(idx):
                item.setSelected(idx == self.tabSpace.currentIndex())


class KSeparator(QFrame):

    """used for toolbar editor"""

    def __init__(self, parent:QWidget) ->None:
        super().__init__(parent)
        self.setLineWidth(1)
        self.setMidLineWidth(0)
        self.setFrameShape(QFrame.Shape.HLine)
        self.setFrameShadow(QFrame.Shadow.Sunken)
        self.setMinimumSize(0, 2)


class ToolBarItem(QListWidgetItem):

    """a toolbar item"""

    def __init__(self, action:Action, parent:QListWidget) ->None:
        self.action = action
        self.parent = parent
        self.emptyIcon:QIcon
        super().__init__(self.__icon(), self.__text(), parent)
        # drop between items, not onto items
        self.setFlags(
            cast(Qt.ItemFlag, (self.flags() | Qt.ItemFlag.ItemIsDragEnabled) & ~Qt.ItemFlag.ItemIsDropEnabled))

    def __icon(self) ->QIcon:
        """the action icon, default is an empty icon"""
        result = self.action.icon()
        if result.isNull():
            if not hasattr(self, 'emptyIcon'):
                if style := self.parent.style():
                    iconSize = style.pixelMetric(
                        QStyle.PixelMetric.PM_SmallIconSize)
                    _ = QPixmap(iconSize, iconSize)
                    _.fill(Qt.GlobalColor.transparent)
                    self.emptyIcon = QIcon(_)
            result = self.emptyIcon
        return result

    def __text(self) ->str:
        """the action text"""
        return self.action.text().replace('&', '')


class ToolBarList(QListWidget):

    """QListWidget without internal moves"""

    def __init__(self, parent:QWidget) ->None:
        super().__init__(parent)
        self.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)  # no internal moves


class KEditToolBar(KDialog):

    """stub"""

    def __init__(self, parent:Optional[QWidget]=None) ->None:
        # pylint: disable=too-many-statements
        super().__init__(parent)
        self.setCaption(i18n('Configure Toolbars'))
        StateSaver(self)
        self.inactiveLabel = QLabel(i18n("A&vailable actions:"), self)
        self.inactiveList = ToolBarList(self)
        self.inactiveList.setDragEnabled(True)
        self.inactiveList.setMinimumSize(180, 250)
        self.inactiveList.setDropIndicatorShown(False)
        self.inactiveLabel.setBuddy(self.inactiveList)
        self.inactiveList.itemSelectionChanged.connect(
            self.inactiveSelectionChanged)
        self.inactiveList.itemDoubleClicked.connect(self.insertButton)
        self.inactiveList.setSortingEnabled(True)

        self.activeLabel = QLabel(i18n('Curr&ent actions:'), self)
        self.activeList = ToolBarList(self)
        self.activeList.setDragEnabled(True)
        self.activeLabel.setBuddy(self.activeList)
        self.activeList.itemSelectionChanged.connect(
            self.activeSelectionChanged)
        self.activeList.itemDoubleClicked.connect(self.removeButton)

        self.upAction = QToolButton(self)
        self.upAction.setIcon(KIcon('go-up'))
        self.upAction.setEnabled(False)
        self.upAction.setAutoRepeat(True)
        self.upAction.clicked.connect(self.upButton)
        self._insertAction:QToolButton = QToolButton(self)
        self._insertAction.setIcon(
            KIcon('go-next' if QApplication.isRightToLeft() else 'go-previous'))
        self._insertAction.setEnabled(False)
        self._insertAction.clicked.connect(self.insertButton)
        self._removeAction:QToolButton = QToolButton(self)
        self._removeAction.setIcon(
            KIcon('go-previous' if QApplication.isRightToLeft() else 'go-next'))
        self._removeAction.setEnabled(False)
        self._removeAction.clicked.connect(self.removeButton)
        self.downAction = QToolButton(self)
        self.downAction.setIcon(KIcon('go-down'))
        self.downAction.setEnabled(False)
        self.downAction.setAutoRepeat(True)
        self.downAction.clicked.connect(self.downButton)

        top_layout = QVBoxLayout(self)
        top_layout.setContentsMargins(0, 0, 0, 0)
        list_layout = QHBoxLayout()

        inactive_layout = QVBoxLayout()
        active_layout = QVBoxLayout()

        button_layout = QGridLayout()

        button_layout.setSpacing(0)
        button_layout.setRowStretch(0, 10)
        button_layout.addWidget(self.upAction, 1, 1)
        button_layout.addWidget(self._removeAction, 2, 0)
        button_layout.addWidget(self._insertAction, 2, 2)
        button_layout.addWidget(self.downAction, 3, 1)
        button_layout.setRowStretch(4, 10)

        inactive_layout.addWidget(self.inactiveLabel)
        inactive_layout.addWidget(self.inactiveList, 1)
        active_layout.addWidget(self.activeLabel)
        active_layout.addWidget(self.activeList, 1)

        list_layout.addLayout(inactive_layout)
        list_layout.addLayout(button_layout)
        list_layout.addLayout(active_layout)

        top_layout.addLayout(list_layout, 10)
        top_layout.addWidget(KSeparator(self))
        self.loadActions()

        self.buttonBox = QDialogButtonBox()
        top_layout.addWidget(self.buttonBox)
        self.setButtons(KDialog.Ok | KDialog.Cancel)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

    def accept(self) ->None:
        """save and close"""
        self.saveActions()
        self.hide()

    def reject(self) ->None:
        """do not save and close"""
        self.hide()

    def inactiveSelectionChanged(self) ->None:
        """update buttons"""
        if self.inactiveList.selectedItems():
            self._insertAction.setEnabled(True)
        else:
            self._insertAction.setEnabled(False)

    def activeSelectionChanged(self) ->None:
        """update buttons"""
        row = self.activeList.currentRow()
        toolItem = None
        if self.activeList.selectedItems():
            toolItem = self.activeList.selectedItems()[0]
        self._removeAction.setEnabled(bool(toolItem))
        if toolItem:
            self.upAction.setEnabled(bool(row))
            self.downAction.setEnabled(row < self.activeList.count() - 1)
        else:
            self.upAction.setEnabled(False)
            self.downAction.setEnabled(False)

    def insertButton(self) ->None:
        """activate an action"""
        self.__moveItem(toActive=True)

    def removeButton(self) ->None:
        """deactivate an action"""
        self.__moveItem(toActive=False)

    def __moveItem(self, toActive:bool) ->None:
        """move item between the two lists"""
        if toActive:
            fromList = self.inactiveList
            toList = self.activeList
        else:
            fromList = self.activeList
            toList = self.inactiveList
        item = cast(ToolBarItem, fromList.takeItem(fromList.currentRow()))
        ToolBarItem(item.action, toList)

    def upButton(self) ->None:
        """move action up"""
        self.__moveUpDown(moveUp=True)

    def downButton(self) ->None:
        """move action down"""
        self.__moveUpDown(moveUp=False)

    def __moveUpDown(self, moveUp:bool) ->None:
        """change place of action in list"""
        active = self.activeList
        row = active.currentRow()
        item = active.takeItem(row)
        offset = -1 if moveUp else 1
        newRow = row + offset
        active.insertItem(newRow, item)
        active.setCurrentRow(newRow)

    def loadActions(self) ->None:
        """load active actions from Preferences"""
        assert Internal.mainWindow
        assert Internal.Preferences
        for name, action in Internal.mainWindow.actionCollection().actions().items():
            if action.text():
                if name in Internal.Preferences.toolBarActions:  # type:ignore
                    ToolBarItem(action, self.activeList)
                else:
                    ToolBarItem(action, self.inactiveList)

    def saveActions(self) ->None:
        """write active actions into Preferences"""
        if Internal.mainWindow is None:
            return
        activeActions = (cast(ToolBarItem, self.activeList.item(
            x)).action for x in range(self.activeList.count()))
        names = {
            v: k for k,
            v in Internal.mainWindow.actionCollection(
            ).actions(
            ).items(
            )}
        assert Internal.Preferences
        Internal.Preferences.toolBarActions = ','.join(
            names[x] for x in activeActions)
        Internal.mainWindow.refreshToolBar()

KGlobal.initStatic()



# -*- coding: utf-8 -*-

"""
 Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import Dict, Type

from common import ReprMixin

class TileSource:
    """
    some constants
    """

    byChar : Dict[str, Type] = {}

    class SourceClass(ReprMixin):
        """Defines defaults"""
        isDiscarded = False
        char: str

        def __str__(self) ->str:
            return self.__repr__()

        def __repr__(self) ->str:
            return 'TileSource.' + self.__class__.__name__.rsplit('.', maxsplit=1)[-1]

    class LivingWallDiscard(SourceClass):
        """Last Tile was discarded"""
        char = 'd'
        isDiscarded = True

    class LivingWall(SourceClass):
        """Last tile comes from wall"""
        char = 'w'

    class East14th(SourceClass):
        """This is the 14th tile for East"""
        char = '1'

    class RobbedKong(SourceClass):
        """Last tile comes from robbing a kong"""
        char = 'k'

    class DeadWall(SourceClass):
        """Last tile comes from dead wall"""
        char = 'e'

    class LivingWallEnd(SourceClass):
        """Last tile comes from living wall and is the last living wall tile"""
        char = 'z'

    class LivingWallEndDiscard(SourceClass):
        """like LivingWallEnd but discarded"""
        char = 'Z'
        isDiscarded = True

    class Unknown(SourceClass):
        """Unknown source"""
        char = '.'

TileSource.byChar['w'] = TileSource.LivingWall
TileSource.byChar['d'] = TileSource.LivingWallDiscard
TileSource.byChar['z'] = TileSource.LivingWallEnd
TileSource.byChar['e'] = TileSource.DeadWall
TileSource.byChar['Z'] = TileSource.LivingWallEndDiscard
TileSource.byChar['1'] = TileSource.East14th
TileSource.byChar['k'] = TileSource.RobbedKong
TileSource.byChar['.'] = TileSource.Unknown



# -*- coding: utf-8 -*-

"""Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only




Read the user manual for a description of the interface to this scoring engine
"""

from itertools import chain
import weakref
from hashlib import md5
from typing import List, Optional, TYPE_CHECKING, Set, Dict, Tuple, Type, Any, Union

from log import dbgIndent
from tile import Tile, TileList, TileTuple, Meld, MeldList
from tilesource import TileSource
from rule import Score, UsedRule
from common import Debug, ReprMixin, num_encode
from util import callers
from message import Message

if TYPE_CHECKING:
    from player import Player
    from tile import Tiles
    from wind import Wind
    from rule import Rule, Ruleset

class Hand(ReprMixin):

    """represent the hand to be evaluated.

    lenOffset is
      <0 for a short hand
      0 for a correct calling hand
      1 for a correct winner hand or a long loser hand
      >1 for a long winner hand
    Of course ignoring bonus tiles and respecting kong replacement tiles.
    if there are no kongs, 13 tiles will return 0

    We assume that long hands never happen. For manual scoring, this should
    be asserted by the caller after creating the Hand instance. If the Hand
    has lenOffset 1 but is no winning hand, the Hand instance will not be
    fully evaluated, it is given Score 0 and hand.won == False.

    declaredMelds are those which cannot be changed anymore: Chows, Pungs,
    Kongs.

    tilesInHand are those not in declaredMelds

    Only tiles passed in the 'R' substring may be rearranged.

    mjRule is the one out of mjRules with the highest resulting score. Every
    hand gets an mjRule even it is not a wining hand, it is the one which
    was used for rearranging the hiden tiles to melds.

    suits include dragons and winds."""

    # pylint: disable=too-many-instance-attributes

    indent = 0
    class __NotWon(UserWarning):  # pylint: disable=invalid-name

        """should be won but is not a winning hand"""

    def __new__(cls, player:'Player', string:Optional[str]=None, melds:Optional[MeldList]=None,  # pylint: disable=too-many-arguments
            unusedTiles:Optional['Tiles']=None, bonusTiles:Optional['Tiles']=None,
            lastSource:Type[TileSource.SourceClass]=TileSource.Unknown, lastTile:Optional[Tile]=None,
            lastMeld:Optional[Meld]=None, announcements:Optional[Set[str]]=None,
            prevHand:Optional['Hand']=None) ->'Hand':
        """since a Hand instance is never changed, we can use a cache"""
        if string:
            cache = player.handCache
            cacheKey = string
            if cacheKey in cache:
                result = cache[cacheKey]
                player.cacheHits += 1
                result.is_from_cache = True
                return result
            player.cacheMisses += 1
            result = object.__new__(cls)
            cache[cacheKey] = result
        else:
            result = object.__new__(cls)
        result.is_from_cache = False
        return result

    def __init__(self, player:'Player', string:Optional[str]=None, melds:Optional[MeldList]=None,  # pylint: disable=too-many-arguments
            unusedTiles:Optional['Tiles']=None, bonusTiles:Optional['Tiles']=None,
            lastSource:Type[TileSource.SourceClass]=TileSource.Unknown, lastTile:Optional[Tile]=None,
            lastMeld:Optional[Meld]=None, announcements:Optional[Set[str]]=None,
            prevHand:Optional['Hand']=None) ->None:
        """evaluate string for player. rules are to be applied in any case"""

        # pylint: disable=too-many-statements
        self.is_from_cache:bool
        if self.is_from_cache:
            return

        # shortcuts for speed:
        self._player = weakref.ref(player)
        assert player.game
        self.ruleset:'Ruleset' = player.game.ruleset
        self.__robbedTile = Tile.none
        self.prevHand = prevHand
        self.__won = None
        self.__score:Score = Score()
        self.__callingHands:Optional[List['Hand']] = None
        self.__mjRule:Optional['Rule'] = None
        self.ruleCache:Dict[Tuple[Type, str], Any] = {}
        self.__lastTile = Tile.none
        self.__lastSource:Type[TileSource.SourceClass] = TileSource.Unknown
        self.__announcements:Set[str] = set()
        self.__lastMeld:Union[Meld, int, None] = 0
        self.__lastMelds = MeldList()
        self.tiles:TileList = TileList()
        self.melds:MeldList = MeldList()
        self.bonusMelds:MeldList = MeldList()
        self.usedRules:List[UsedRule] = []
        self.unusedTiles:TileList = TileList()
        self.__arranged:Optional[bool] = None
        self.lenOffset:int

        if string:
            self.__parseString(string)
        else:
            self.melds = melds or MeldList()
            if unusedTiles is not None:
                self.unusedTiles.extend(unusedTiles)  # FIXME: assign
            if bonusTiles:
                self.bonusMelds = MeldList(Meld(x) for x in bonusTiles)
            self.__lastSource = lastSource
            if lastTile:
                self.__lastTile = lastTile
            if lastMeld:
                self.__lastMeld= lastMeld
            self.__announcements = announcements or set()
            # FIXME: TileList() should suffice, but it does not yet resolve melds to tiles. TileTuple does.
            self.tiles = TileList(TileTuple(chain(self.melds, self.unusedTiles)))
            string = self.newString()
        self.string:str = string

        self.__precompute()

        self.__won = self.lenOffset == 1 and player.mayWin

        if Debug.hand or (Debug.mahJongg and self.lenOffset == 1):
            self.debug(f'{callers(exclude=["__init__"])}')
            Hand.indent += 1
            self.debug(f'New Hand {string} lenOffset={self.lenOffset}')

        try:
            self.__arrange()
            self.__calculate()
            self.__arranged = True
        except Hand.__NotWon as notwon:
            if Debug.mahJongg:
                self.debug(str(notwon))
            self.__won = False
            self.__score = Score()
        finally:
            self._fixed = True
            if Debug.hand or (Debug.mahJongg and self.lenOffset == 1):
                self.debug(f"Fixing {self} {'won ' if self.won else ''}{self.score}")
            Hand.indent -= 1

    def __parseString(self, inString:str) ->None:
        """parse the string passed to Hand()"""
        tileStrings = []
        for part in inString.split():
            partId = part[0]
            if partId == 'm':
                if len(part) > 1:
                    try:
                        self.__lastSource = TileSource.byChar[part[1]]
                    except KeyError as _:
                        raise KeyError(f'{inString} has unknown lastTile {part[1]}') from _
                    if len(part) > 2:
                        self.__announcements = set(part[2])
            elif partId == 'L':
                if len(part[1:]) > 8:
                    raise ValueError(
                        'last tile cannot complete a kang:' + inString)
                if len(part) > 3:
                    self.__lastMeld = Meld(part[3:])
                self.__lastTile = Tile(part[1:3])
            else:
                if part != 'R':
                    tileStrings.append(part)
        self.bonusMelds, tileStrings = self.__separateBonusMelds(tileStrings)
        tileString = ' '.join(tileStrings)
        self.tiles = TileList(tileString.replace(' ', '').replace('R', ''))
        for part in tileStrings[:]:
            if part[:1] != 'R':
                self.melds.append(Meld(part))
                tileStrings.remove(part)
        assert len(tileStrings) < 2, tileStrings
        if tileStrings:
            self.unusedTiles.extend(TileList(tileStrings[0][1:]))

    def __precompute(self) ->None:
        """precompute commonly used things"""
        self.tiles = TileList(self.tiles.sorted())

        self.values = tuple(x.value for x in self.tiles)
        self.suits = {x.lowerGroup for x in self.tiles}
        self.declaredMelds = MeldList(x for x in self.melds if x.isDeclared)
        declaredTiles = TileTuple(self.declaredMelds)
        self.tilesInHand = TileList(x for x in self.tiles
                                    if x not in declaredTiles)
        self.lenOffset = (len(self.tiles) - self.ruleset.dealtTiles
                          - sum(x.isKong for x in self.melds))

        last = self.__lastTile
        if last and last.isKnown and not last.isBonus:
#            print('lastTile %s hand.tiles %s, string=%s hand %s' % (last, self.tiles, self.string, str(self)))
            assert last in self.tiles, \
                f'lastTile {last} is not in hand.tiles {self.tiles}, hand {str(self)}'
            if self.__lastSource is TileSource.RobbedKong:
                assert self.tiles.count(last.exposed) + \
                    self.tiles.count(last.concealed) == 1, (
                        f"Robbing kong: I cannot have lastTile {last} more than once in {' '.join(self.tiles)}")
        self.newStr = self.newString()

    @property
    def arranged(self) ->Optional[bool]:
        """readonly"""
        return self.__arranged

    @property
    def player(self) ->'Player':
        """weakref"""
        result = self._player()
        assert result
        return result

    @property
    def ownWind(self) ->'Wind':
        """for easier usage"""
        return self.player.wind

    @property
    def roundWind(self) ->'Wind':
        """for easier usage"""
        assert self.player.game
        return self.player.game.point.prevailing

    def __calculate(self) ->None:
        """apply rules, calculate score"""
        # TODO: in __init__: self.__score == self.__calculate()
        assert not self.unusedTiles, (
            f'Hand.__calculate expects there to be no unused tiles: {self}')
        oldWon = self.__won
        self.__applyRules()
        if len(self.lastMelds) > 1:
            self.__applyBestLastMeld()
        if self.__won != oldWon:
            # if not won after all, this might be a long hand.
            # So we might even have to unapply meld rules and
            # bonus points. Instead just recompute all again.
            # This should only happen with scoring manual games
            # and with scoringtest - normally kajongg would not
            # let you declare an invalid mah jongg
            self.__applyRules()
        if not self.__score:
            self.__score = Score()

    def hasTiles(self) ->bool:
        """tiles are assigned to this hand"""
        return bool(self.tiles or self.bonusMelds)

    @property
    def mjRule(self) ->Optional['Rule']:
        """getter"""
        return self.__mjRule

    @mjRule.setter
    def mjRule(self, value:'Rule') ->None:
        """changing mjRule must reset score"""
        if self.__mjRule != value:
            self.__mjRule = value
            self.__score = Score()

    @property
    def lastTile(self) -> Tile:
        """compute and cache, readonly"""
        return self.__lastTile

    @property
    def lastSource(self) ->Type[TileSource.SourceClass]:
        """compute and cache, readonly"""
        return self.__lastSource

    @property
    def announcements(self) ->Set[str]:
        """compute and cache, readonly"""
        return self.__announcements

    @property
    def score(self) ->Score:
        """calculate it first if not yet done"""
        if self.__score is None and self.__arranged is not None:
            self.__calculate()
        assert self.__score is not None, f'cannot calculate score for {self}'
        return self.__score

    @property
    def lastMeld(self) ->Optional[Meld]:
        """compute and cache, readonly"""
        if self.__lastMeld == 0:
            self.__setLastMeld()
        return self.__lastMeld  # type:ignore[return-value]

    @property
    def lastMelds(self) ->MeldList:
        """compute and cache, readonly"""
        if self.__lastMeld == 0:
            self.__setLastMeld()
        return self.__lastMelds

    @property
    def won(self) ->bool:
        """do we really have a winner hand?"""
        return bool(self.__won)

    def debug(self, msg:str) ->None:
        """try to use Game.debug so we get a nice prefix"""
        idPrefix = num_encode(hash(self))
        if self.prevHand:
            idPrefix += f'<{num_encode(hash(self.prevHand))}'
        idPrefix = f'Hand({idPrefix})'
        assert self.player.game
        self.player.game.debug(' '.join([dbgIndent(self, self.prevHand), idPrefix, msg]))

    def __applyRules(self) ->None:
        """find out which rules apply, collect in self.usedRules"""
        self.usedRules = []
        for meld in chain(self.melds, self.bonusMelds):
            self.usedRules.extend(UsedRule(x, meld) for x in meld.rules(self))
        for rule in self.ruleset.handRules:
            if rule.appliesToHand(self):
                self.usedRules.append(UsedRule(rule))

        self.__score = self.__totalScore()

        self.ruleCache.clear()
        # do the rest only if we know all tiles of the hand
        if self.string:
            if Tile.unknownStr in self.string:
                return
        else:
# FIXME: try to do this also for string
            if Tile.unknown in self.tiles:
                return
        if self.__won:
            matchingMJRules = self.__maybeMahjongg()
            if not matchingMJRules:
                self.__score = Score()
                raise Hand.__NotWon('no matching MJ Rule')
            self.__mjRule = matchingMJRules[0]
            self.usedRules.append(UsedRule(self.__mjRule))
            self.usedRules.extend(self.matchingWinnerRules())
            self.__score = self.__totalScore()
        else:  # not self.won
            loserRules = self.__matchingRules(self.ruleset.loserRules)
            if loserRules:
                self.usedRules.extend(UsedRule(x) for x in loserRules)
                self.__score = self.__totalScore()
        self.__checkHasExclusiveRules()

    def matchingWinnerRules(self) ->List[UsedRule]:
        """return a list of matching winner rules"""
        matching = [UsedRule(x) for x in self.__matchingRules(self.ruleset.winnerRules)]
        limitRule = self.maxLimitRule(matching)
        return [limitRule] if limitRule else matching

    def __checkHasExclusiveRules(self) ->None:
        """if we have one, remove all others"""
        exclusive = [x for x in self.usedRules if 'absolute' in x.rule.options]
        if exclusive:
            self.usedRules = exclusive
            self.__score = self.__totalScore()
            if self.__won and not bool(self.__maybeMahjongg()):
                raise Hand.__NotWon(f'exclusive rule {exclusive} does not win')

    def __setLastMeld(self) ->None:
        """set the shortest possible last meld. This is
        not yet the final choice, see __applyBestLastMeld"""
        self.__lastMeld = None
        if self.lastTile and self.__won:
            if self.mjRule:
                self.__lastMelds = self.mjRule.computeLastMelds(self)  # type:ignore[call-arg,misc,attr-defined]
                if self.__lastMelds:
                    # syncHandBoard may return nothing
                    if len(self.__lastMelds) == 1:
                        self.__lastMeld = self.__lastMelds[0]
                    else:
                        totals = sorted(
                            (len(x), idx)
                            for idx, x in enumerate(self.__lastMelds))
                        self.__lastMeld = self.__lastMelds[totals[0][1]]
            if not self.__lastMeld:
                self.__lastMeld = self.lastTile.single
                self.__lastMelds = MeldList(self.__lastMeld)

    def __applyBestLastMeld(self) ->None:
        """select the last meld giving the highest score
        (only winning variants)"""
        assert len(self.lastMelds) > 1
        totals = []
        prev = self.lastMeld
        for rule in self.usedRules:
            assert isinstance(rule, UsedRule)
        for lastMeld in self.lastMelds:
            self.__lastMeld = lastMeld
            try:
                self.__applyRules()
                totals.append((self.__totalScore().total(), lastMeld))
            except Hand.__NotWon:
                pass
        if totals:
            totals = sorted(totals)  # sort by totalScore
            maxScore = totals[-1][0]
            bestLastMelds = [x[1] for x in totals if x[0] == maxScore]
            # now we have a list of only lastMelds reaching maximum score
            if prev not in bestLastMelds or self.__lastMeld not in bestLastMelds:
                if Debug.explain and prev not in bestLastMelds:
                    assert self.player.game
                    if not self.player.game.belongsToRobotPlayer():
                        self.debug(f'replaced last meld {prev} with {bestLastMelds[0]}')
                self.__lastMeld = bestLastMelds[0]
                self.__applyRules()

    def chancesToWin(self) ->List[Tile]:
        """count the physical tiles that make us win and still seem available"""
        assert self.lenOffset == 0
        result = []
        for completedHand in self.callingHands:
            assert completedHand.lastTile
            result.extend(
                [completedHand.lastTile] *
                (self.player.tileAvailable(completedHand.lastTile, self)))
        return result

    def assertEqual(self, other:'Hand') ->None:
        """raise assertion if not equal with detailled info"""
        assert self.melds == other.melds, \
            f'Melds in hands differ:{self.melds!r} != {other.melds!r}'
        assert self.lastTile == other.lastTile, \
            f'lastTile in hands differs:{self.lastTile} != {other.lastTile}'
        assert self.lastMeld == other.lastMeld, \
            f'lastMeld in hands differs:{self.lastMeld} != {other.lastMeld}'
        assert self.newString() == other.newString(), \
            f'newString() in hands differs:{self.newString()} != {other.newString()}'
        assert self.newStr == other.newStr, \
            f'newStr in hands differs:{self.newStr} != {other.newStr}'

    def newString(self, melds:MeldList=1, unusedTiles:Optional['Tiles']=1,  # type:ignore[assignment]
        lastSource:Optional[Type[TileSource.SourceClass]]=1, announcements:Set[str]=1,  # type:ignore[assignment]
        lastTile:Tile=1, lastMeld:Optional[Meld]=1) ->str:  # type:ignore[assignment]  # type:ignore[assignment]
        """create string representing a hand. Default is current Hand, but every part
        can be overridden or excluded by passing None"""
        if melds == 1:
            melds = MeldList(chain(self.melds, self.bonusMelds))
        if unusedTiles== 1:
            unusedTiles = self.unusedTiles
        if lastSource == 1:
            lastSource = self.lastSource
        if announcements == 1:
            announcements = self.announcements
        if lastTile == 1:
            lastTile = self.lastTile
        if lastMeld == 1:
            lastMeld = self.__lastMeld  # type:ignore[assignment]
        parts = [str(x) for x in sorted(melds)]
        if unusedTiles:
            parts.append('R' + ''.join(str(x) for x in sorted(unusedTiles)))
        if lastSource or announcements:
            parts.append(f"m{self.lastSource.char}{''.join(self.announcements)}")
        if lastTile:
            parts.append(f"L{lastTile}{lastMeld if lastMeld else ''}")
            self.player.assertLastTile()
        return ' '.join(parts).strip()

    def __add__(self, addTile:Tile) ->'Hand':
        """return a new Hand built from this one plus addTile"""
        assert addTile.isConcealed, f'addTile {addTile} should be concealed:'
        # combine all parts about hidden tiles plus the new one to one part
        # because something like DrDrS8S9 plus S7 will have to be reordered
        # anyway
        newString = self.newString(
            melds=MeldList(chain(self.declaredMelds, self.bonusMelds)),
            unusedTiles=self.tilesInHand + [addTile],
            lastSource=None,
            lastTile=addTile,
            lastMeld=None
            )
        return Hand(self.player, newString, prevHand=self)

    def __sub__(self, subtractTile:Tile) ->'Hand':
        """return a copy of self minus subtractTiles.
        Case of subtractTile (hidden or exposed) is ignored.
        subtractTile must either be undeclared or part of
        lastMeld. Exposed melds of length<3 will be hidden."""
        # If lastMeld is given, it must be first in the list.
        # Next try undeclared melds, then declared melds
        assert self.lenOffset == 1, \
             f'lenOffset != 1: {self.lenOffset} for {self}'  # correct winner hand or long loser hand
        if self.lastTile:
            if self.lastTile is subtractTile and self.prevHand:
                return self.prevHand
        declaredMelds = self.declaredMelds
        tilesInHand = TileList(self.tilesInHand)
        boni = MeldList(self.bonusMelds)
        lastMeld = self.lastMeld
        if subtractTile.isBonus:
            for idx, meld in enumerate(boni):
                if subtractTile is meld[0]:
                    del boni[idx]
                    break
        else:
            if lastMeld and lastMeld.isDeclared and (
                    subtractTile.exposed in lastMeld.exposed):
                try:
                    declaredMelds.remove(lastMeld)
                except ValueError as _:
                    raise ValueError(
                        f'lastMeld {lastMeld} is not in declaredMelds {declaredMelds}, hand is: {self}') from _
                tilesInHand.extend(lastMeld.concealed)
            try:
                tilesInHand.remove(subtractTile.concealed)
            except ValueError as _:
                raise ValueError(
                f'subtractTile.concealed={subtractTile.concealed} is not in tilesInHand {tilesInHand}') from _

        for meld in declaredMelds[:]:
            if len(meld) < 3:
                declaredMelds.remove(meld)
                tilesInHand.extend(meld.concealed)
        # if we robbed a kong, remove that announcement
        mjPart = ''
        announcements = self.announcements - set('k')
        if announcements:
            mjPart = 'm.' + ''.join(announcements)
        rest = 'R' + str(tilesInHand)
        newString = ' '.join(str(x) for x in (
            declaredMelds, rest, boni, mjPart))
        return Hand(self.player, newString, prevHand=self)

    def manualRuleMayApply(self, rule:'Rule') ->bool:
        """return True if rule has selectable() and applies to this hand"""
        if self.__won and rule in self.ruleset.loserRules:
            return False
        if not self.__won and rule in self.ruleset.winnerRules:
            return False
        return rule.selectable(self) or rule.appliesToHand(self)
        # needed for activated rules

    @property
    def callingHands(self) ->List['Hand']:
        """the hand is calling if it only needs one tile for mah jongg.
        Returns all hands which would only need one tile.
        If mustBeAvailable is True, make sure the missing tile might still
        be available.
        """
        if self.__callingHands is None:
            self.__callingHands = self.__findAllCallingHands()
        return self.__callingHands

    def __findAllCallingHands(self) ->List['Hand']:
        """always try to find all of them"""
        result:List['Hand'] = []
        if self.lenOffset:
            return result
        candidates:TileList = TileList()
        for rule in self.ruleset.mjRules:
            cand = rule.winningTileCandidates(self)
            if Debug.hand and cand:
                candis = ''.join(str(x) for x in sorted(cand))
                self.debug(f'callingHands found {candis} for {rule}')
            candidates.extend(x.concealed for x in cand)
        for tile in sorted(set(candidates)):
            if sum(x.exposed == tile.exposed for x in self.tiles) == 4:
                continue
            hand = self + tile
            if hand.won:
                result.append(hand)
        if Debug.hand:
            _hiderules = ', '.join({x.mjRule.name for x in result if x.mjRule})
            if _hiderules:
                self.debug(f'Is calling {_hiderules}')
        return result

    @property
    def robbedTile(self) ->Tile:
        """cache this here for use in rulecode"""
        if self.__robbedTile is Tile.unknown:
            self.__robbedTile = Tile.none
            assert self.player.game
            if self.player.game.moves:
                # scoringtest does not (yet) simulate this
                lastMove = self.player.game.moves[-1]
                if (lastMove.message == Message.DeclaredKong
                        and lastMove.player != self.player):
                    self.__robbedTile = lastMove.meld[1]
                    # we want it concealed only for a hidden Kong
        return self.__robbedTile

    def __maybeMahjongg(self) ->List['Rule']:
        """check if this is a mah jongg hand.
        Return a sorted list of matching MJ rules, highest
        total first."""
        if self.lenOffset == 1 and self.player.mayWin:
            matchingMJRules = [x for x in self.ruleset.mjRules
                               if x.appliesToHand(self)]
            if matchingMJRules:
                if self.robbedTile and self.robbedTile.isConcealed:
                    # Millington 58: robbing hidden kong is only
                    # allowed for 13 orphans
                    matchingMJRules = [
                        x for x in matchingMJRules
                        if 'mayrobhiddenkong' in x.options]
                result = sorted(matchingMJRules, key=lambda x: -x.score.total())
                if Debug.mahJongg:
                    self.debug(f'{callers()} Found {matchingMJRules}')
                return result
        return []

    def __arrangements(self) ->List[Tuple['Rule', MeldList]]:
        """find all legal arrangements.
        Returns a list of tuples with the mjRule and a list of concealed melds"""
        self.unusedTiles.sort()
        result:List[Tuple['Rule', MeldList]] = []
        stdMJ = self.ruleset.standardMJRule
        assert stdMJ
        if self.mjRule:
            rules = [self.mjRule]
        else:
            rules = self.ruleset.mjRules
        for mjRule in rules:
            if ((self.lenOffset == 1 and mjRule.appliesToHand(self))
                    or (self.lenOffset < 1 and mjRule.shouldTry(self))):  # type:ignore[attr-defined]
                if self.unusedTiles:
                    unused = TileList(Tile(x) for x in self.unusedTiles)
                    for melds, rest2 in mjRule.rearrange(self, unused):  # type:ignore[attr-defined]
                        if rest2:
                            melds = MeldList(melds)
                            restMelds, _ = next(
                                stdMJ.rearrange(self, rest2[:]))  # type:ignore[attr-defined]
                            melds.extend(restMelds)
                        result.append((mjRule, melds))
        if not result:
            result.extend(
                (stdMJ, x[0])
                for x in stdMJ.rearrange(self, self.unusedTiles[:]))  # type:ignore[attr-defined]
        return result

    def __arrange(self) ->None:
        """work hard to always return the variant with the highest Mah Jongg value."""
        if any(not x.isKnown for x in self.unusedTiles):
            meldCount, restCount = divmod(len(self.unusedTiles), 3)
            assert Tile.unknown.pung
            self.melds.extend([Tile.unknown.pung] * meldCount)
            if restCount:
                self.melds.append(Meld(Tile.unknown * restCount))
            self.unusedTiles = TileList()
        if not self.unusedTiles:
            self.melds.sort()
            mjRules = self.__maybeMahjongg()
            if self.won:
                if not mjRules:
                    # how could this ever happen?
                    raise Hand.__NotWon('Long Hand with no unused tiles')
                self.mjRule = mjRules[0]
            return
        wonHands = []
        lostHands = []
        for mjRule, melds in self.__arrangements():
            allMelds = self.melds[:] + list(melds)
            lastTile = self.lastTile
            if self.lastSource and self.lastSource.isDiscarded:
                lastTile = lastTile.exposed
                lastMelds = sorted(
                    (x for x in allMelds if not x.isDeclared and lastTile.concealed in x),
                    key=lambda x: len(x)) # pylint: disable=unnecessary-lambda
                if lastMelds:
                    allMelds.remove(lastMelds[0])
                    allMelds.append(lastMelds[0].exposed)
            _ = self.newString(
                MeldList(chain(allMelds, self.bonusMelds)),
                unusedTiles=None, lastTile=lastTile, lastMeld=None)
            tryHand = Hand(self.player, _, prevHand=self)
            if tryHand.won:
                tryHand.mjRule = mjRule
                wonHands.append((mjRule, melds, tryHand))
            else:
                lostHands.append((mjRule, melds, tryHand))
        # we prefer a won Hand even if a lost Hand might have a higher score
        tryHands = wonHands if wonHands else lostHands
        bestRule, bestVariant, _ = max(tryHands, key=lambda x: x[2])
        if wonHands:
            self.mjRule = bestRule
        self.melds.extend(bestVariant)
        self.melds.sort()
        self.unusedTiles = TileList()
        self.ruleCache.clear()

    def __gt__(self, other:Any) ->bool:
        """compares hand values"""
        assert self.player == other.player
        if not other.arranged:
            return True
        if self.won and not (other.arranged and other.won):
            return True
        if not (self.arranged and self.won) and other.won:
            return False
        _ = self.player.intelligence
        return _.handValue(self) > _.handValue(other)

    def __lt__(self, other:Any) ->bool:
        """compares hand values"""
        return other.__gt__(self)

    def __eq__(self, other:Any) ->bool:
        """compares hand values"""
        assert self.__class__ is other.__class__, \
            f'Hands have different classes:{self.__class__} and {other.__class__}'
        assert self.player == other.player
        return self.newStr == other.newStr

    def __ne__(self, other:Any) ->bool:
        """compares hand values"""
        assert self.__class__ is other.__class__
        assert self.player == other.player
        return self.newStr != other.newStr

    def __matchingRules(self, rules:List['Rule']) ->List['Rule']:
        """return all matching rules for this hand"""
        return [rule for rule in rules if rule.appliesToHand(self)]

    @staticmethod
    def maxLimitRule(usedRules:List[UsedRule]) ->Optional[UsedRule]:
        """return the rule with the highest limit score or None"""
        result = None
        maxLimit = 0.0
        usedRules = [x for x in usedRules if x.rule.score.limits]
        for usedRule in usedRules:
            score = usedRule.rule.score
            if score.limits > maxLimit:
                maxLimit = score.limits
                result = usedRule
        return result

    def __totalScore(self) ->Score:
        """use all used rules to compute the score"""
        maxRule = self.maxLimitRule(self.usedRules)
        maxLimit = 0.0
        pointsTotal = sum((x.rule.score for x in self.usedRules),
                          Score(ruleset=self.ruleset))
        if maxRule:
            maxLimit = maxRule.rule.score.limits
            if (maxLimit >= 1.0
                    or maxLimit * self.ruleset.limit > pointsTotal.total()):
                self.usedRules = [maxRule]
                return Score(ruleset=self.ruleset, limits=maxLimit)
        return pointsTotal

    def total(self) ->int:
        """total points of hand"""
        return self.score.total()

    @staticmethod
    def __separateBonusMelds(tileStrings:List[str]) ->Tuple[MeldList, List[str]]:
        """One meld per bonus tile. Others depend on that."""
        bonusMelds = MeldList()
        for tileString in tileStrings[:]:
            if len(tileString) == 2:
                tile = Tile(tileString)
                if tile.isBonus:
                    bonusMelds.append(tile.single)
                    tileStrings.remove(tileString)
        return bonusMelds, tileStrings

    def explain(self) ->List[str]:
        """explain what rules were used for this hand"""
        usedRules = self.player.sortRulesByX(self.usedRules)
        result = [x.rule.explain(x.meld) for x in usedRules
                  if x.rule.score.points]
        result.extend(
            [x.rule.explain(x.meld) for x in usedRules
             if x.rule.score.doubles])
        result.extend(
            [x.rule.explain(x.meld) for x in usedRules
             if not x.rule.score.points and not x.rule.score.doubles])
        if any(x.rule.debug for x in usedRules):
            result.append(str(self))
        return result

    def doublesEstimate(self, discard:Optional[Tile]=None) ->int:
        """this is only an estimate because it only uses meldRules and handRules,
        but not things like mjRules, winnerRules, loserRules"""
        result = 0
        if discard and self.tiles.count(discard) == 2:
            melds = chain(self.melds, self.bonusMelds, [discard.exposed.pung])
        else:
            melds = chain(self.melds, self.bonusMelds)
        for meld in melds:
            result += sum(x.score.doubles for x in meld.doublingRules(self))
        for rule in self.ruleset.doublingHandRules:
            if rule.appliesToHand(self):
                result += rule.score.doubles
        return result

    def __str__(self) ->str:
        """hand as a string"""
        return self.newString()

    def __hash__(self) ->int:
        """used for debug logging to identify the hand"""
        if not hasattr(self, 'string'):
            return 0
        md5sum = md5()
        md5sum.update(self.player.name.encode('utf-8'))
        _ = self.string if self.string else str(self)
        md5sum.update(_.encode())
        digest = md5sum.digest()
        assert len(digest) == 16
        result = 0
        for part in range(4):
            result = (result << 8) + digest[part]
        return result



# -*- coding: utf-8 -*-

"""
Copyright (C) 2010-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import functools
import types

from typing import List, Any, Optional, Dict, Union, TYPE_CHECKING, Callable, Type, cast

from twisted.internet.defer import Deferred, succeed, fail

from qt import QPropertyAnimation, QParallelAnimationGroup, \
    QAbstractAnimation, QEasingCurve
from qt import Property, QGraphicsObject, QGraphicsItem, QPointF, QObject

from common import Internal, Debug, isAlive, ReprMixin, id4
from log import logDebug, logFailure

if TYPE_CHECKING:
    from qt import QGraphicsScene

PropertyType = Union[QPointF,int,float]

class Animation(QPropertyAnimation, ReprMixin):

    """a Qt animation with helper methods"""

    nextAnimations : List['Animation'] = []
    clsUid = 0

    def __init__(self, graphicsObject:'AnimatedMixin', propName:str,
        endValue:PropertyType, parent:Optional['QObject']=None) ->None:
        self.debug = graphicsObject.debug_name() in Debug.animation or Debug.animation == 'all'
        self.debug |= f'T{id4(graphicsObject)}t' in Debug.animation
        Animation.clsUid += 1
        self.uid = Animation.clsUid
        assert isinstance(graphicsObject, QObject)
        _ = propName.encode()
        super().__init__(graphicsObject, _, parent)
        super().setEndValue(endValue)
        assert Internal.Preferences
        duration = Internal.Preferences.animationDuration()
        self.setDuration(duration)
        self.setEasingCurve(QEasingCurve.Type.InOutQuad)
        graphicsObject.queuedAnimations.append(self)
        Animation.nextAnimations.append(self)
        if self.debug:
            oldAnimation = graphicsObject.activeAnimation.get(propName, None)
            if isAlive(oldAnimation):
                assert isinstance(oldAnimation, Animation)
                logDebug(
                    f'new Animation({self}) (after {oldAnimation.ident()} is done)')
            else:
                logDebug(f'Animation({self})')

    def setEndValue(self, endValue:PropertyType) ->None:
        """wrapper with debugging code"""
        graphicsObject = self.targetObject()
        if not isAlive(graphicsObject):
            # may happen when aborting a game because animations are cancelled first,
            # before the last move from server is executed
            return
        if cast('AnimatedMixin', graphicsObject).debug_name() in Debug.animation or Debug.animation == 'all':
            logDebug(
                f'{self.ident()}: change endValue for {self.pName()}: '
                f'{self.formatValue(self.endValue())}->{self.formatValue(endValue)}  {graphicsObject}')
        QPropertyAnimation.setEndValue(self, endValue)

    def ident(self) ->str:
        """the identifier to be used in debug messages"""
        pGroup = self.group() if isAlive(self) else 'notAlive'
        if pGroup or not isAlive(self):
            return f'{pGroup}/A_{id4(self)}'
        return f"A_{id4(self)}-{cast('AnimatedMixin', self.targetObject()).debug_name()}"

    def pName(self) ->str:
        """
        Return self.propertyName() as a python string.

        @return: C{str}
        """
        if not isAlive(self):
            return 'notAlive'
        _ = self.propertyName()
        return bytes(_).decode()

    def formatValue(self, value:PropertyType) ->str:
        """string format the wanted value from qvariant"""
        pName = self.pName()
        if pName == 'pos':
            _ = cast(QPointF, value)
            return f'{_.x():.0f}/{_.y():.0f}'
        if pName == 'rotation':
            _ = cast(int, value)
            return str(_)
        if pName == 'scale':
            _ = cast(float, value)
            return f'{_:.2f}'
        return f'formatValue: unexpected {pName}={value}'

    def __str__(self) ->str:
        """for debug messages"""
        if not isAlive(self):
            return 'notAlive'
        if not isAlive(self.targetObject()):
            return f'{self.ident()} {self.pName()}: target notAlive'
        currentValue = getattr(self.targetObject(), self.pName())
        endValue = self.endValue()
        targetObject = None
        if _ := self.targetObject():
            targetObject = _
        return (f'{self.ident()} {self.pName()}: {self.formatValue(currentValue)}->'
                f'{self.formatValue(endValue)} for {targetObject} duration={int(self.duration())}ms')

    @staticmethod
    def removeImmediateAnimations() ->None:
        """execute and remove immediate moves from the list
        We do not animate objects if
             - we are in a graphics object drag/drop operation
             - the user disabled animation
             - there are too many animations in the group so it would be too slow
             - the object has duration 0
        """
        if Animation.nextAnimations:
            needRefresh = False
            assert Internal.mainWindow
            assert Internal.Preferences
            shortcutAll = (Internal.scene is None
                           or Internal.mainWindow.centralView.dragObject
                           or Internal.Preferences.animationSpeed == 99
                           or len(Animation.nextAnimations) > 1000)
                    # change 1000 to 100 if we do not want to animate shuffling and
                    # initial deal
            for animation in Animation.nextAnimations[:]:
                if shortcutAll or animation.duration() == 0:
                    cast(AnimatedMixin, animation.targetObject()).shortcutAnimation(animation)
                    Animation.nextAnimations.remove(animation)
                    needRefresh = True
            if needRefresh and Internal.scene:
                Internal.scene.focusRect.refresh()


class ParallelAnimationGroup(QParallelAnimationGroup, ReprMixin):

    """
    current is the currently executed group
    doAfter is a list of Deferred to be called when this group
    is done. If another group is chained to this one, transfer
    doAfter to that other group.
    """

    running : List['ParallelAnimationGroup'] = []  # we need a reference to active animation groups
    current = None
    clsUid = 0
    def __init__(self, animations:List[Animation], parent:Optional['QObject']=None) ->None:
        super().__init__(parent)
        self.animations = animations
        self.uid = ParallelAnimationGroup.clsUid
        ParallelAnimationGroup.clsUid += 1
        self.deferred:Deferred = Deferred()
        self.deferred.addErrback(logFailure)
        self.steps = 0
        self.debug = any(x.debug for x in self.animations)
        self.debug |= f'G{id4(self)}g' in Debug.animation
        self.doAfter:List[Deferred] = []
        if ParallelAnimationGroup.current:
            if self.debug or ParallelAnimationGroup.current.debug:
                logDebug(f'Chaining Animation group G_{id4(self)} to G{ParallelAnimationGroup.current}')
            self.doAfter = ParallelAnimationGroup.current.doAfter
            ParallelAnimationGroup.current.doAfter = []
            ParallelAnimationGroup.current.deferred.addCallback(self.start).addErrback(logFailure)
        else:
            self.start()
        ParallelAnimationGroup.running.append(self)
        ParallelAnimationGroup.current = self
        self.stateChanged.connect(self.showState)

    @staticmethod
    def cancelAll() ->None:
        """cancel all animations"""
        if Debug.quit:
            logDebug('Cancelling all animations')
        for group in ParallelAnimationGroup.running:
            if isAlive(group):
                group.clear()

    def showState(self, newState:int, oldState:int) ->None:
        """override Qt method"""
        if self.debug:
            logDebug(f'G{self.uid}: {self.stateName(oldState)} -> {self.stateName(newState)} isAlive:{isAlive(self)}')

    def updateCurrentTime(self, value:int) ->None:
        """count how many steps an animation does."""
        self.steps += 1
        if self.steps % 50 == 0:
            # periodically check if the board still exists.
            # if not (game end), we do not want to go on
            for animation in self.animations:
                graphicsObject = cast(AnimatedMixin, animation.targetObject())
                if hasattr(graphicsObject, 'board') and not isAlive(graphicsObject.board):
                    graphicsObject.clearActiveAnimation(animation)
                    self.removeAnimation(animation)
        super().updateCurrentTime(value)

    def start(self, unusedResults:str='DIREKT') ->Deferred:  # type: ignore[override]
        """start the animation, returning its deferred"""
        if not isAlive(self):
            return fail()
        assert self.state() != QAbstractAnimation.State.Running
        for animation in self.animations:
            graphicsObject = animation.targetObject()
            if not isAlive(animation) or not isAlive(graphicsObject):
                return fail()
            animatedObject = cast(AnimatedMixin, graphicsObject)
            animatedObject.setActiveAnimation(animation)
            self.addAnimation(animation)
            propName = animation.pName()
            animation.setStartValue(animatedObject.getValue(propName))
            if propName == 'rotation':
                # change direction if that makes the difference smaller
                endValue = animation.endValue()
                currValue = cast(float, animatedObject.rotation)
                if endValue - currValue > 180:
                    animation.setStartValue(currValue + 360)
                if currValue - endValue > 180:
                    animation.setStartValue(currValue - 360)
        for animation in self.animations:
            if target := animation.targetObject():
                target.setDrawingOrder()  # type:ignore[attr-defined]
        self.finished.connect(self.allFinished)
        scene = Internal.scene
        assert scene
        scene.focusRect.hide()
        QParallelAnimationGroup.start(
            self,
            QAbstractAnimation.DeletionPolicy.DeleteWhenStopped)
        if self.debug:
            assert Internal.Preferences
            _ = ','.join(f'A_{id4(x)}' for x in self.animations)
            logDebug(f'{self} started with speed {int(Internal.Preferences.animationSpeed)} ({_})')
        return succeed(None).addErrback(logFailure)

    def allFinished(self) ->None:
        """all animations have finished. Cleanup and callback"""
        self.fixAllBoards()
        if self == ParallelAnimationGroup.current:
            ParallelAnimationGroup.current = None
            ParallelAnimationGroup.running = []
        if Debug.animationSpeed and self.duration():
            perSecond = self.steps * 1000.0 / self.duration()
            if perSecond < 50:
                logDebug(f'{int(self.steps)} steps for {len(self.children())} animations, {perSecond:.1f}/sec')
        # if we have a deferred, callback now
        assert self.deferred
        if self.debug:
            logDebug(f'Done: {self}')
        if self.deferred:
            self.deferred.callback(None)
        for after in self.doAfter:
            after.callback(None)

    def fixAllBoards(self) ->None:
        """set correct drawing order for all moved graphics objects"""
        for animation in self.children():
            if graphicsObject := animation.targetObject():  # type:ignore[attr-defined]
                graphicsObject.clearActiveAnimation(animation)
        if Internal.scene:
            Internal.scene.focusRect.refresh()

    def stateName(self, state:Any=None) ->str:
        """for debug output"""
        if not isAlive(self):
            return 'not alive'
        if state is None:
            state = self.state()
        if state == QAbstractAnimation.State.Stopped:
            return 'stopped'
        if state == QAbstractAnimation.State.Running:
            return 'running'
        return f'unknown state:{state}'

    def __str__(self) ->str:
        """for debugging"""
        return f'G{self.uid}({len(self.animations)}:{self.stateName()})'


class AnimatedMixin:
    """for UITile and WindDisc"""

    def __init__(self) ->None:
        super().__init__()
        self.activeAnimation:Dict[str, Animation]  = {}  # key is the property name
        self.queuedAnimations:List[Animation] = []

    def _get_pos(self) ->QPointF:
        """getter for property pos"""
        return QGraphicsObject.pos(cast(QGraphicsItem, self))

    def _set_pos(self, pos:QPointF) ->None:
        """setter for property pos"""
        QGraphicsObject.setPos(cast(QGraphicsItem, self), pos)

    pos = Property(QPointF, fget=_get_pos, fset=_set_pos)

    def _get_scale(self) ->float:
        """getter for property scale"""
        return QGraphicsObject.scale(cast(QGraphicsItem, self))

    def _set_scale(self, scale:float) ->None:
        """setter for property scale"""
        QGraphicsObject.setScale(cast(QGraphicsItem, self), scale)

    scale = Property(float, fget=_get_scale, fset=_set_scale)

    def _get_rotation(self) ->float:
        """getter for property rotation"""
        return QGraphicsObject.rotation(cast(QGraphicsItem, self))

    def _set_rotation(self, rotation:float) ->None:
        """setter for property rotation"""
        QGraphicsObject.setRotation(cast(QGraphicsItem, self), rotation)

    rotation = Property(float, fget=_get_rotation, fset=_set_rotation)

    def queuedAnimation(self, propertyName:str) ->Optional['Animation']:
        """return the last queued animation for this graphics object and propertyName"""
        for item in reversed(self.queuedAnimations):
            if item.pName() == propertyName:
                return item
        return None

    def debug_name(self) ->str:
        """for mypy"""
        return ''

#    this results in player names appearing BELOW their walls
#    def setDrawingOrder(self) ->None:
#        """for mypy"""
#

    def shortcutAnimation(self, animation:'Animation') ->None:
        """directly set the end value of the animation"""
        if animation.debug:
            logDebug(f'shortcut {animation}: UTile {self.debug_name()}: clear queuedAnimations')
        setattr(self, animation.pName(), animation.endValue())
        self.queuedAnimations = []
        self.setDrawingOrder()  # type:ignore[attr-defined] # TODO: mypy protocol?

    def getValue(self, pName:str) ->Union[QPointF,int,float]:
        """get a current property value"""
        if pName == 'pos':
            return cast(QPointF, self.pos)
        if pName == 'rotation':
            return cast(int, self.rotation)
        if pName == 'scale':
            return cast(float, self.scale)
        assert False

    def setActiveAnimation(self, animation:'Animation') ->None:
        """the graphics object knows which of its properties are currently animated"""
        self.queuedAnimations = []
        propName = animation.pName()
        if self.debug_name() in Debug.animation:
            oldAnimation = self.activeAnimation.get(propName, None)
            if not isAlive(oldAnimation):
                oldAnimation = None
            if oldAnimation:
                logDebug(f'**** setActiveAnimation {self.debug_name()} {propName}: '
                         f'{animation} OVERRIDES {oldAnimation}')
            else:
                logDebug(f'setActiveAnimation {self.debug_name()} {propName}: set {animation}')
        self.activeAnimation[propName] = animation
        self.setCacheMode(QGraphicsItem.CacheMode.ItemCoordinateCache)  # type: ignore[attr-defined]

    def clearActiveAnimation(self, animation:'Animation') ->None:
        """an animation for this graphics object has ended.
        Finalize graphics object in its new position"""
        del self.activeAnimation[animation.pName()]
        if self.debug_name() in Debug.animation:
            logDebug(f'UITile {self.debug_name()}: clear activeAnimation_{animation.pName()}')
        self.setDrawingOrder()  # type:ignore[attr-defined] # TODO: mypy protocol?
        if not self.activeAnimation:
            self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)  # type: ignore[attr-defined]
            self.update()  # type: ignore[attr-defined]

    def setupAnimations(self) ->None:
        """move the item to its new place. This puts new Animation
        objects into the queue to be animated by calling animate()"""
        for pName, newValue in self.moveDict().items():  # type: ignore[attr-defined]
            if self.scene() != Internal.scene:  # type: ignore[attr-defined]
                # not part of the playing scene, like tiles in tilesetselector
                setattr(self, pName, newValue)
                continue
            animation = self.queuedAnimation(pName)
            if animation:
                curValue = animation.endValue()
                if curValue != newValue:
                    # change a queued animation
                    if self.debug_name() in Debug.animation:
                        logDebug(f'setEndValue for {animation}: {pName}: '
                                 f'{animation.formatValue(curValue)}->{animation.formatValue(newValue)}')
                    animation.setEndValue(newValue)
            else:
                animation = self.activeAnimation.get(pName, None)
                if isAlive(animation):
                    assert isinstance(animation, Animation)
                    curValue = animation.endValue()
                else:
                    curValue = self.getValue(pName)
                if pName != 'scale' or abs(curValue - newValue) > 0.00001:
                    # ignore rounding differences for scale
                    if curValue != newValue:
                        Animation(self, pName, newValue)


class AnimationSpeed:

    """a helper class for moving graphics with a given speed. 99=immediate."""

    def __init__(self, speed:Optional[int]=None) ->None:
        if speed is None:
            speed = 99
        if Internal.Preferences:
            self.__speed = speed
            self.prevAnimationSpeed = Internal.Preferences.animationSpeed
            if Internal.Preferences.animationSpeed != speed:
                Internal.Preferences.animationSpeed = speed
                if Debug.animationSpeed:
                    logDebug(f'AnimationSpeed sets speed {int(speed)}')

    def __enter__(self) ->'AnimationSpeed':
        return self

    def __exit__(self, exc_type:Type[Exception], exc_value:Exception, trback:Any) ->None:
        """reset previous animation speed"""
        if Internal.Preferences:
            if self.__speed < 99:
                animate()
            if Internal.Preferences.animationSpeed != self.prevAnimationSpeed:
                if Debug.animationSpeed:
                    logDebug(f'AnimationSpeed restores speed '
                             f'{Internal.Preferences.animationSpeed} to {self.prevAnimationSpeed}')
                Internal.Preferences.animationSpeed = self.prevAnimationSpeed


def afterQueuedAnimations(doAfter:Deferred) ->Callable:
    """A decorator"""

    @functools.wraps(doAfter)  # type:ignore[arg-type]
    def doAfterQueuedAnimations(*args:Any, **kwargs:Any) ->None:
        """do this after all queued animations have finished"""
        method = types.MethodType(doAfter, args[0])  # type:ignore[arg-type]
        args = args[1:]
        varnames = doAfter.__code__.co_varnames  # type:ignore[attr-defined]
        assert varnames[1] in ('deferredResult', 'unusedDeferredResult'), \
            f'{doAfter.__qualname__} passed {varnames[1]} instead of deferredResult'  # type:ignore[attr-defined]
        animateAndDo(method, *args, **kwargs)

    return doAfterQueuedAnimations


def animate() ->Deferred:
    """now run all prepared animations. Returns a Deferred
        so callers can attach callbacks to be executed when
        animation is over.
    """
    if Animation.nextAnimations:
        Animation.removeImmediateAnimations()
        animations = Animation.nextAnimations
        if animations:
            Animation.nextAnimations = []
            return ParallelAnimationGroup(animations).deferred
    elif ParallelAnimationGroup.current:
        return ParallelAnimationGroup.current.deferred
    return succeed(None).addErrback(logFailure)


def doCallbackWithSpeed(result:Any, speed:int, callback:Callable, *args:Any, **kwargs:Any) ->None:
    """as the name says"""
    with AnimationSpeed(speed):
        callback(result, *args, **kwargs)


def animateAndDo(callback:Callable, *args:Any, **kwargs:Any) ->Deferred:
    """if we want the next animations to have the same speed as the current group,
    do not use animate().addCallback() because speed would not be kept"""
    result = animate()
    if Internal.Preferences:
        # we might be called very early
        speed = Internal.Preferences.animationSpeed
        assert isinstance(speed, int)
        result.addCallback(doCallbackWithSpeed, speed, callback, *args, **kwargs)
        result.addErrback(logFailure)
    return result



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""
from typing import List, TYPE_CHECKING, Optional, Literal, Dict, Any, Union, cast, Sequence

from common import Internal, ZValues, ReprMixin, Speeds, DrawOnTopMixin
from wind import Wind, East
from qt import QPointF, QGraphicsObject, QFontMetrics, QGraphicsItem
from qt import QPen, QColor, QFont, QRectF

from guiutil import Painter, sceneRotation
from board import WindDisc, YellowText, Board
from wall import Wall, KongBox
from tile import Tile
from tileset import Tileset
from uitile import UITile
from animation import animate, afterQueuedAnimations, AnimationSpeed
from animation import ParallelAnimationGroup, AnimatedMixin, animateAndDo

if TYPE_CHECKING:
    from twisted.internet.defer import Deferred
    from qt import QPainter, QStyleOptionGraphicsItem, QWidget
    from tile import Piece
    from visible import VisiblePlayingGame
    from scoring import ScoringGame
    from game import Game
    from scene import PlayingScene


class SideText(AnimatedMixin, QGraphicsObject, ReprMixin, DrawOnTopMixin): # type:ignore[misc]

    """The text written on the wall"""

    sideTexts : List['SideText'] = []

    def __init__(self, parent:Optional[QGraphicsItem]=None) ->None:
        assert parent is None
        assert len(self.sideTexts) < 4
        self.__name = f't{len(self.sideTexts)}'
        self.sideTexts.append(self)
        super().__init__()
        self.hide()
        assert Internal.scene
        Internal.scene.addItem(self)
        self.__text = ''
        self.__board:Optional['UIWallSide'] = None
        self.needsRefresh = False
        self.__color = QColor('black')
        self.__boundingRect:Optional[QRectF] = None
        self.__font:QFont = QFont()
        self.animateNextChange = False
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsFocusable, False)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable, False)

    def adaptedFont(self) ->QFont:
        """Font with the correct point size for the wall"""
        result = QFont()
        size = 80
        result.setPointSize(size)
        assert self.__board
        tileHeight = self.__board.tileset.faceSize.height()
        while QFontMetrics(result).ascent() > tileHeight:
            size -= 1
            result.setPointSize(size)
        return result

    @staticmethod
    def refreshAll() ->None:
        """recompute ourself. Always do this for all for sides
        together because if two names change place we want the
        to move simultaneously"""
        sides = SideText.sideTexts
        if all(not x.needsRefresh for x in sides):
            return
        rotating = False
        for side in sides:
            side.show()
            if not side.needsRefresh:
                continue
            side.needsRefresh = False
            assert side.board
            rotating |= sceneRotation(side) != sceneRotation(side.board)

        alreadyMoved = any(x.x() for x in sides)
        speed = 99
        for side in sides:
            if side.animateNextChange:
                speed = Speeds.windDisc if rotating and alreadyMoved else 99
                side.animateNextChange = False
        with AnimationSpeed(speed=speed):
            # first round: just place the winds. Only animate moving them
            # for later rounds.
            for side in sides:
                side.setupAnimations()
        animate()

    @staticmethod
    def removeAll() ->None:
        """from the scene"""
        assert Internal.scene
        for side in SideText.sideTexts:
            Internal.scene.removeItem(side)
        SideText.sideTexts = []

    def moveDict(self) -> Dict[Literal['pos', 'rotation', 'scale'], Any]:
        """return a dict with new property values for our sidetext
        which move it onto us"""
        if not self.board or not self.__text:
            return {}
        rotation = sceneRotation(self.board)
        position = self.board.center()
        textCenter = self.boundingRect().center()
        if rotation == 180:
            rotation = 0
            position += textCenter
        else:
            position -= textCenter
        return {'pos': self.board.mapToScene(position), 'rotation': rotation, 'scale': self.board.scale()}

    def debug_name(self) ->str:
        """for identification in animations"""
        return self.__name

    @property
    def board(self) ->Optional['UIWallSide']:
        """the front we are sitting on"""
        return self.__board

    @board.setter
    def board(self, value:Optional['UIWallSide']) ->None:
        if self.__board != value:
            self.__board = value
            self.__font = self.adaptedFont()
            self.needsRefresh = True

    @property
    def color(self) ->QColor:
        """text color"""
        return self.__color

    @color.setter
    def color(self, value:QColor) ->None:
        if self.__color != value:
            self.__color = value
            self.update()

    @property
    def text(self) ->str:
        """what we are saying"""
        return self.__text

    @text.setter
    def text(self, value:str) ->None:
        if self.__text != value:
            self.__text = value
            self.prepareGeometryChange()
            txt = self.__text
            if ' - ' in txt:
                # this disables animated movement if only the score changes
                txt = txt[:txt.rfind(' - ')] + ' - 55'
            self.__boundingRect = QRectF(QFontMetrics(self.__font).boundingRect(txt))
            self.needsRefresh = True

    def paint(self, painter:Optional['QPainter'], unusedOption:Optional['QStyleOptionGraphicsItem'],
        unusedWidget:Optional['QWidget']=None) ->None:
        """paint the disc"""
        if painter:
            with Painter(painter):
                pen = QPen(self.color)
                painter.setPen(pen)
                painter.setFont(self.__font)
                painter.drawText(0, 0, self.__text)

    def boundingRect(self) ->QRectF:
        """around the text"""
        return self.__boundingRect or QRectF()

    def __str__(self) ->str:
        """for debugging"""
        return f'SideText({self.debug_name()} {self.text} x/y= {self.x():.1f}/{self.y():1f})'


class UIWallSide(Board, ReprMixin):

    """a Board representing a wall of tiles"""
    penColor = QColor('red')

    def __init__(self, tileset:Tileset, boardRotation:int, length:float):
        super().__init__(length, 1, tileset, boardRotation=boardRotation)
        self.length = length
        self.disc:'WindDisc'
        self.message:YellowText

    def debug_name(self) ->str:
        """name for debug messages"""
        return f'UIWallSide {UIWall.sideNames[int(self.rotation())]}'

    def center(self) ->QPointF:
        """return the center point of the wall in relation to the
        faces of the upper level"""
        faceRect = self.tileFaceRect()
        result = faceRect.topLeft()
        result += self.shiftZ(1)
        result += QPointF(self.length // 2 * faceRect.width(), faceRect.height() / 2)
        result.setX(result.x() + faceRect.height() / 2)  # corner tile
        return result

    def hide(self) ->None:
        """hide all my parts"""
        self.disc.hide()
        Board.hide(self)

    def __str__(self) ->str:
        """for debugging"""
        return self.debug_name()


class UIKongBox(KongBox):

    """Kong box with UITiles"""

    def __init__(self) ->None:
        KongBox.__init__(self)

    def fill(self, tiles:List[Union['Piece', UITile]]) ->None:
        """fill the box"""
        for uiTile in self._tiles:
            cast(UITile, uiTile).cross = False
        super().fill(tiles)
        for uiTile in self._tiles:
            cast(UITile, uiTile).cross = True

    def pop(self, count:int) ->List[Union['Piece', UITile]]:
        """get count tiles from kong box"""
        result = super().pop(count)
        for uiTile in result:
            cast(UITile, uiTile).cross = False
        return result


class UIWall(Wall):

    """represents the wall with four sides. self.wall[] indexes them
    counter clockwise, 0..3. 0 is bottom."""

    Lower, Right, Upper, Left = range(4)
    sideAngles = (0, 270, 180, 90)
    sideNames = {0:'Lower', 1:'Right', 2:'Upper', 3:'Left'}
    sideNames[270] = 'Right'
    sideNames[180] = 'Upper'
    sideNames[90] = 'Left'

    tileClass = UITile
    kongBoxClass = UIKongBox

    def __init__(self, game:'Game') ->None:
        """init and position the wall"""
        # we use only white dragons for building the wall. We could actually
        # use any tile because the face is never shown anyway.
        self.initWindDiscs()
        game.wall = self
        super().__init__(game)
        self.tiles:Sequence[UITile] = [cast(UITile, x) for x in self.tiles]  # type:ignore[assignment]
        self.__square = Board(1, 1, Tileset.current())
        self.__square.setZValue(ZValues.markerZ)
        sideLength = len(self.tiles) // 8
        self.__sides = [UIWallSide(
            Tileset.current(),
            boardRotation, sideLength) for boardRotation in self.sideAngles]
        for idx, side in enumerate(self.__sides):
            side.setParentItem(self.__square)
            side.lightSource = self.lightSource
            side.disc = Wind.all4[idx].disc
            side.disc.hide()
            side.message = YellowText(side)
            side.message.setZValue(ZValues.popupZ)
            side.message.setVisible(False)
            side.message.setPos(side.center())
        self.__sides[self.Lower].setTilePos(yWidth=sideLength)
        self.__sides[self.Left].setTilePos(xHeight=1)
        self.__sides[self.Upper].setTilePos(xHeight=1, xWidth=sideLength, yHeight=1)
        self.__sides[self.Right].setTilePos(xWidth=sideLength, yWidth=sideLength, yHeight=1)
        assert Internal.scene
        Internal.scene.addItem(self.__square)
        assert Internal.Preferences
        Internal.Preferences.addWatch('showShadows', self.showShadowsChanged)

    @staticmethod
    def initWindDiscs() ->None:
        """the 4 round wind discs on the player walls"""
        assert Internal.scene
        if not hasattr(East, 'disc'):
            for wind in Wind.all4:
                wind.disc = WindDisc(wind)
                Internal.scene.addItem(wind.disc)

    @staticmethod
    def debug_name() ->str:
        """name for debug messages"""
        return 'wall'

    def __getitem__(self, index:int) ->'UIWallSide':
        """make Wall index-able"""
        return self.__sides[index]

    def __setitem__(self, index:int, value:'UIWallSide') ->None:
        """only for pylint, currently not used"""
        self.__sides[index] = value

    def __delitem__(self, index:int) ->None:
        """only for pylint, currently not used"""
        del self.__sides[index]

    def __len__(self) ->int:
        """only for pylint, currently not used"""
        return len(self.__sides)

    def hide(self) ->None:
        """hide all four walls and their decorators"""
        # may be called twice
        self.living = []
        self.kongBox.fill([])
        for side in self.__sides:
            side.hide()
        self.tiles = []
        if scene := self.__square.scene():
            scene.removeItem(self.__square)

    def __shuffleTiles(self) ->None:
        """shuffle tiles for next hand"""
        assert Internal.scene
        discardBoard = cast('PlayingScene', Internal.scene).discardBoard
        places = [(x, y) for x in range(-3, int(discardBoard.width) + 3)
                  for y in range(-3, int(discardBoard.height) + 3)]
        assert self.game
        places = self.game.randomGenerator.sample(places, len(self.tiles))
        for idx, uiTile in enumerate(self.tiles):
            uiTile.dark = True
            uiTile.setBoard(discardBoard, *places[idx])

    def build(self, shuffleFirst:bool=False) ->'Deferred':
        """builds the wall without dividing"""
        # recycle used tiles
        for uiTile in self.tiles:
            uiTile.change_name(Tile.unknown)
            uiTile.dark = True
        with AnimationSpeed(speed=99):
            # FIXME: introduce Speeds.shuffle = 90, but then tiles appear in wrong places
            if shuffleFirst:
                self.__shuffleTiles()
            for uiTile in self.tiles:
                uiTile.focusable = False
            return animateAndDo(self.__placeWallTiles)

    def __placeWallTiles(self, unusedResult:Any=None) ->'Deferred':
        """place all wall tiles"""
        tileIter = iter(self.tiles)
        tilesPerSide = len(self.tiles) // 4
        for side in (self.__sides[0], self.__sides[3],
                     self.__sides[2], self.__sides[1]):
            upper = True  # upper tile is played first
            for position in range(tilesPerSide - 1, -1, -1):
                uiTile = next(tileIter)
                uiTile.setBoard(side, position // 2, 0, level=int(upper))
                upper = not upper
        self.__setDrawingOrder()
        return animate()

    @property
    def lightSource(self) ->Union[Literal['NE'], Literal['NW'], Literal['SW'], Literal['SE']]:
        """see LIGHTSOURCES"""
        return cast(Union[Literal['NE'], Literal['NW'], Literal['SW'], Literal['SE']], self.__square.lightSource)

    @lightSource.setter
    def lightSource(self, lightSource:Union[Literal['NE'], Literal['NW'], Literal['SW'], Literal['SE']]) ->None:
        """setting this actually changes the visuals"""
        if self.lightSource != lightSource:
#            assert ParallelAnimationGroup.current is None # may trigger, reason unknown
            self.__square.lightSource = lightSource
            for side in self.__sides:
                side.lightSource = lightSource
            self.__setDrawingOrder()
            SideText.refreshAll()

    @property
    def tileset(self) ->Tileset:
        """The tileset of this wall"""
        return self.__square.tileset

    @tileset.setter
    def tileset(self, value:Tileset) ->None:
        """setting this actually changes the visuals."""
        if self.tileset != value:
            assert ParallelAnimationGroup.current is None
            self.__square.tileset = value
            self.__resizeHandBoards()
            SideText.refreshAll()

    @afterQueuedAnimations  # type:ignore[arg-type]
    def showShadowsChanged(self, deferredResult:Any, # pylint: disable=unused-argument
        unusedOldValue:bool, unusedNewValue:bool) ->None:
        """setting this actually changes the visuals."""
        self.__resizeHandBoards()

    def __resizeHandBoards(self, unusedResults:Any=None) ->None:
        """we are really calling _setRect() too often. But at least it works"""
        assert self.game
        for player in self.game.players:
            player.handBoard.computeRect()
        assert Internal.mainWindow
        Internal.mainWindow.adjustMainView()

    def __setDrawingOrder(self, unusedResults:Any=None) ->None:
        """set drawing order of the wall"""
        levels = {'NW': (2, 3, 1, 0), 'NE': (
            3, 1, 0, 2), 'SE': (1, 0, 2, 3), 'SW': (0, 2, 3, 1)}
        for idx, side in enumerate(self.__sides):
            side.level = (
                levels[
                    side.lightSource][
                        idx] + 1) * ZValues.boardZFactor

    def __moveDividedTile(self, uiTile:UITile, offset:float) ->None:
        """moves a uiTile from the divide hole to its new place"""
        board = uiTile.board
        newOffset = uiTile.xoffset + offset
        sideLength = len(self.tiles) // 8
        if newOffset >= sideLength:
            assert board
            sideIdx = self.__sides.index(cast(UIWallSide, board))
            board = self.__sides[(sideIdx + 1) % 4]
        uiTile.setBoard(board, newOffset % sideLength, 0, level=2)
        uiTile.update()

    @afterQueuedAnimations  # type:ignore[arg-type]
    def _placeLooseTiles(self, deferredResult:Any=None) ->None:
        """place the last 2 tiles on top of kong box"""
        assert len(self.kongBox) % 2 == 0
        placeCount = len(self.kongBox) // 2
        if placeCount >= 4:
            first = min(placeCount - 1, 5)
            second = max(first - 2, 1)
            self.__moveDividedTile(cast(UITile, self.kongBox[-1]), second)
            self.__moveDividedTile(cast(UITile, self.kongBox[-2]), first)

    def divide(self) ->None:
        """divides a wall, building a living end and a dead end"""
        with AnimationSpeed():
            Wall.divide(self)
            for uiTile in self.tiles:
                # update graphics because tiles having been
                # in kongbox in a previous game
                # might not be there anymore. This gets rid
                # of the cross on them.
                # FIXME: Piece/UITile
                uiTile.update()  # type:ignore[union_attr]
            # move last two tiles onto the dead end:
            return self._placeLooseTiles()

    def decorate4(self, deferredResult:Any=None) ->None:
        """show player info on the wall. The caller must ensure
        all are moved simultaneously and at which speed by using
        AnimationSpeed.
        already queued animations keep their speed, only the windDiscs
        are moved without animation.
        """
        assert self.game
        with AnimationSpeed(99):
            for player in self.game.players:
                player.showInfo()
            SideText.refreshAll()
        animateAndDo(self.showWindDiscs)

    def showWindDiscs(self, unusedDeferred:Optional['Deferred']=None) ->None:
        """animate all windDiscs. The caller must ensure
        all are moved simultaneously and at which speed
        by using AnimationSpeed."""
        assert self.game
        for player in self.game.players:
            side = player.front
            side.disc.setupAnimations()
            side.disc.show()



"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

partially based on C++ code from:
Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import cast
from qt import QWidget, QLineEdit
from qt import QSize, QMetaObject, QFormLayout, QVBoxLayout, QHBoxLayout, QSizePolicy
from qt import QLabel, QGroupBox, QListWidget, QFrame, QSpacerItem
from background import Background
from common import Internal
from mi18n import i18nc


class BackgroundSelector(QWidget):

    """presents all available backgrounds with previews"""

    # pylint: disable=too-many-instance-attributes,too-many-statements

    def __init__(self, parent:QWidget) ->None:
        super().__init__(parent)
        self.setupUi()
        self.kcfg_backgroundName = QLineEdit(self)
        self.kcfg_backgroundName.setVisible(False)
        self.kcfg_backgroundName.setObjectName('kcfg_backgroundName')
        self.backgroundNameList:'QListWidget'
        self.backgroundAuthor:'QLabel'
        self.backgroundContact:'QLabel'
        self.backgroundDescription:'QLabel'
        self.backgroundPreview:'QLabel'
        self.setUp()

    def setupUi(self) ->None:
        """created by pyuic from old backgroundselector.ui and adapted for direct use"""
        self.setObjectName("BackgroundSelector")
        self.resize(497, 446)
        self.setMaximumSize(QSize(800, 600))
        self.vboxlayout = QVBoxLayout(self)
        self.vboxlayout.setContentsMargins(0, 0, 0, 0)
        self.vboxlayout.setObjectName("vboxlayout")
        self.hboxlayout = QHBoxLayout()
        self.hboxlayout.setContentsMargins(0, 0, 0, 0)
        self.hboxlayout.setObjectName("hboxlayout")
        self.backgroundNameList = QListWidget(self)
        self.backgroundNameList.setMinimumSize(QSize(120, 0))
        self.backgroundNameList.setObjectName("backgroundNameList")
        self.hboxlayout.addWidget(self.backgroundNameList)
        self.groupBox_2 = QGroupBox(self)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setObjectName("groupBox_2")
        self.vboxlayout1 = QVBoxLayout(self.groupBox_2)
        self.vboxlayout1.setContentsMargins(0, 0, 0, 0)
        self.vboxlayout1.setObjectName("vboxlayout1")
        self.backgroundPreview = QLabel(self.groupBox_2)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.backgroundPreview.sizePolicy().hasHeightForWidth())
        self.backgroundPreview.setSizePolicy(sizePolicy)
        self.backgroundPreview.setMinimumSize(QSize(200, 160))
        self.backgroundPreview.setFrameShape(QFrame.Shape.Box)
        self.backgroundPreview.setObjectName("backgroundPreview")
        self.vboxlayout1.addWidget(self.backgroundPreview)
        self.hboxlayout.addWidget(self.groupBox_2)
        self.vboxlayout.addLayout(self.hboxlayout)
        self.groupBox = QGroupBox(self)
        self.groupBox.setObjectName("groupBox")
        self.formLayout = QFormLayout(self.groupBox)
        self.formLayout.setObjectName("formLayout")
        self.labelAuthor = QLabel(self.groupBox)
        self.labelAuthor.setObjectName("labelAuthor")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.labelAuthor)
        self.backgroundAuthor = QLabel(self.groupBox)
        self.backgroundAuthor.setObjectName("backgroundAuthor")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.backgroundAuthor)
        self.labelContact = QLabel(self.groupBox)
        self.labelContact.setObjectName("labelContact")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.labelContact)
        self.backgroundContact = QLabel(self.groupBox)
        self.backgroundContact.setObjectName("backgroundContact")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.backgroundContact)
        self.labelDescription = QLabel(self.groupBox)
        self.labelDescription.setObjectName("labelDescription")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.LabelRole, self.labelDescription)
        self.backgroundDescription = QLabel(self.groupBox)
        self.backgroundDescription.setObjectName("backgroundDescription")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.backgroundDescription)
        self.vboxlayout.addWidget(self.groupBox)
        spacerItem = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.vboxlayout.addItem(spacerItem)

        self.retranslateUi()
        QMetaObject.connectSlotsByName(self)

    def retranslateUi(self) ->None:
        """created by pyuic from old backgroundselector.ui and adapted for direct use"""
        self.groupBox_2.setTitle(i18nc("BackgroundSelector", "Preview"))
        self.groupBox.setTitle(i18nc("BackgroundSelector", "Properties"))
        self.labelAuthor.setText(i18nc("BackgroundSelector", "Author:"))
        self.labelContact.setText(i18nc("BackgroundSelector", "Contact:"))
        self.labelDescription.setText(i18nc("BackgroundSelector", "Description:"))

    def setUp(self) ->None:
        """fill the selector"""

        # The lineEdit widget holds our background path, but the user does
        # not manipulate it directly
        self.kcfg_backgroundName.hide()

        self.backgroundNameList.currentRowChanged.connect(
            self.backgroundRowChanged)
        self.kcfg_backgroundName.textChanged.connect(
            self.backgroundNameChanged)
        self.backgroundList = Background.available()
        for aset in self.backgroundList:
            self.backgroundNameList.addItem(aset.name)
        assert Internal.Preferences
        self.kcfg_backgroundName.setText(str(Internal.Preferences.backgroundName))

    def backgroundNameChanged(self, name:str) ->None:
        """the name changed: update the current row"""
        igrindex = 0
        for idx, aset in enumerate(self.backgroundList):
            if aset.desktopFileName == name:
                igrindex = idx
                break
        self.backgroundNameList.setCurrentRow(igrindex)

    def backgroundRowChanged(self) ->None:
        """user selected a new background, update our information about it and paint preview"""
        selBackground = cast(Background, self.backgroundList[
            self.backgroundNameList.currentRow()])
        self.kcfg_backgroundName.setText(selBackground.desktopFileName)
        self.backgroundAuthor.setText(selBackground.author)
        self.backgroundContact.setText(selBackground.authorEmail)
        self.backgroundDescription.setText(selBackground.description)
        selBackground.setPalette(self.backgroundPreview)
        self.backgroundPreview.setAutoFillBackground(True)



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

# util must not import from log because util should not
# depend on kde.py

import traceback
import os
import datetime
import subprocess
import gc


from locale import getpreferredencoding
from sys import stdout
from typing import Optional, List, Generator, Sequence, Any, Type, Iterable, Dict, TYPE_CHECKING

from common import Debug

if TYPE_CHECKING:
    from types import FrameType

STDOUTENCODING: Optional[str]

try:
    STDOUTENCODING = stdout.encoding
except AttributeError:
    STDOUTENCODING = None
if not STDOUTENCODING:
    STDOUTENCODING = getpreferredencoding()


def stack(msg:str, limit:int=6) ->List[str]:
    """return a list of lines with msg as prefix"""
    result = []
    for idx, values in enumerate(
            traceback.extract_stack(limit=limit + 2)[:-2]):
        fileName, line, function, txt = values
        result.append(
            f'{int(idx):2}: {msg} {os.path.splitext(os.path.basename(fileName))[0]}/{int(line)} {function}: {txt}')
    return result


def callers(count:int=5, exclude:Optional[Sequence]=None, frame:Optional['FrameType']=None) ->str:
    """return the name of the calling method"""
    stck = traceback.extract_stack(f=frame, limit=30)
    excluding = list(exclude) if exclude else []
    excluding.extend(['<genexpr>', '__call__', 'run', '<module>', 'runTests'])
    excluding.extend(['_startRunCallbacks', '_runCallbacks', 'remote_move', 'exec_move'])
    excluding.extend(['proto_message', '_recvMessage', 'remoteMessageReceived'])
    excluding.extend(['clientAction', 'myAction', 'expressionReceived'])
    excluding.extend(['_read', 'callWithLogger'])
    excluding.extend(['callbackIfDone', 'callback', '__gotAnswer'])
    excluding.extend(['callExpressionReceived', 'proto_answer'])
    excluding.extend(['_dataReceived', 'dataReceived', 'gotItem'])
    excluding.extend(['callWithContext', '_doReadOrWrite', 'doRead'])
    excluding.extend(['callers', 'debug', 'logMessage', 'logDebug'])
    _ = list(x[2] for x in stck if x[2] not in excluding)
    names = reversed(_[-count:])
    result = '.'.join(names)
    return f'[{result}]'


def elapsedSince(since:Optional[datetime.datetime]) ->float:
    """return seconds since since"""
    if not since:
        return 0.0
    delta = datetime.datetime.now() - since
    return float(
        delta.microseconds
        + (delta.seconds + delta.days * 24 * 3600) * 10 ** 6) / 10 ** 6


def which(program:str)->Optional[str]:
    """return the full path for the binary or None"""
    for path in os.environ['PATH'].split(os.pathsep):
        fullName = os.path.join(path, program)
        if os.path.exists(fullName):
            return fullName
    return None


def removeIfExists(filename:str) ->bool:
    """remove file if it exists. Returns True if it existed"""
    exists = os.path.exists(filename)
    if exists:
        os.remove(filename)
    return exists


def uniqueList(seq:Iterable) ->List:
    """makes list content unique, keeping only the first occurrence"""
    seen:Any = set()
    seen_add = seen.add
    return [x for x in seq if x not in seen and not seen_add(x)]


def _getr(slist:Any, olist:Any, seen:Any) ->None:
    """Recursively expand slist's objects into olist, using seen to track
    already processed objects."""
    for element in slist:
        if id(element) in seen:
            continue
        seen[id(element)] = None
        olist.append(element)
        tlist = gc.get_referents(element)
        if tlist:
            _getr(tlist, olist, seen)

# The public function.


def get_all_objects() ->Any:
    """Return a list of all live Python objects, not including the
    list itself. May use this in Duration for showing where
    objects are leaking"""
    gc.collect()
    gcl = gc.get_objects()
    olist:List[Any] = []
    seen:Dict[int, Any] = {}
    # Just in case:
    seen[id(gcl)] = None
    seen[id(olist)] = None
    seen[id(seen)] = None
    # _getr does the real work.
    _getr(gcl, olist, seen)
    return olist


class Duration:

    """a helper class for checking code execution duration"""

    def __init__(self, name:str, threshold:float=1.0, bug:bool=False) ->None:
        """name describes where in the source we are checking
        threshold in seconds: do not warn below
        if bug is True, throw an exception if threshold is exceeded"""
        self.name = name
        self.threshold = threshold
        self.bug = bug
        self.__start = datetime.datetime.now()

    def __enter__(self) ->'Duration':
        return self

    def __exit__(self, exc_type:Type, exc_value:Exception, trback:Any) ->None:
        """now check time passed"""
        if not Debug.neutral:
            diff = datetime.datetime.now() - self.__start
            if diff > datetime.timedelta(seconds=self.threshold):
                msg = f'{self.name} took {int(diff.seconds)}.{int(diff.microseconds):02} seconds'
                if self.bug:
                    raise UserWarning(msg)
                print(msg)

def __debugCollect() ->None:
    """collect using DEBUG_LEAK"""
    gc.set_threshold(0)
    gc.set_debug(gc.DEBUG_LEAK)
    gc.enable()
    print('collecting {{{')
    gc.collect()        # we want to eliminate all output
    print('}}} done')

def checkMemory() ->None:
    """as the name says"""
    if not Debug.gc:
        return

    __debugCollect()

    # code like this may help to find specific things
    if True: # pylint: disable=using-constant-test
        interesting = ('Client', 'Player', 'Game')
        for obj in gc.garbage:
            if hasattr(obj, 'cell_contents'):
                obj = obj.cell_contents
            if not any(x in repr(obj) for x in interesting):
                continue
            for referrer in gc.get_referrers(obj):
                if referrer is gc.garbage:
                    continue
                if hasattr(referrer, 'cell_contents'):
                    referrer = referrer.cell_contents
                if referrer.__class__.__name__ in interesting:
                    for referent in gc.get_referents(referrer):
                        print(f'{referrer} refers to {referent}')
                else:
                    print(f'referrer of {type(obj)}/{obj} is: id={id(referrer)} type={type(referrer)} {referrer}')
    print(f'unreachable:{gc.collect()}')
    gc.set_debug(0)


def gitHead() ->Optional[str]:
    """the current git commit. 'current' if there are uncommitted changes
    and None if no .git found"""
    if not os.path.exists(os.path.join('..', '.git')):
        return None
    subprocess.Popen(['git', 'update-index', '-q', '--refresh'])  # pylint:disable=consider-using-with
    uncommitted = list(popenReadlines('git diff-index --name-only HEAD --'))
    return 'current' if uncommitted else next(popenReadlines('git log -1 --format=%h'))


def popenReadlines(args:Any) ->Generator[str, None, None]:
    """runs a subprocess and returns stdout as a list of unicode encodes lines"""
    if isinstance(args, str):
        args = args.split()
    my_env = os.environ.copy()
    my_env["LANG"] = "C"
    result = subprocess.Popen(args, universal_newlines=True, stdout=subprocess.PIPE, env=my_env).communicate()[0]  # pylint:disable=consider-using-with
    return (x.strip() for x in result.split('\n') if x.strip())
